##
# File: RepoScanExec.py
#
#  Execution wrapper  --  repository scanning utilities --
#
#  Updates:
#
# 28-Jun-2018 jdw update ScanRepoUtil prototype with workPath
#  3-Jul-2018 jdw update to latest ScanRepoUtil() prototype
# 20-Aug-2018 jdw engage incremental repository scan mode.
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import argparse
import logging
import os
import sys
from collections import OrderedDict

from rcsb.db.define.DictionaryApiProviderWrapper import DictionaryApiProviderWrapper
from rcsb.db.utils.ScanRepoUtil import ScanRepoUtil
from rcsb.utils.config.ConfigUtil import ConfigUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


def scanRepo(
    cfgOb,
    contentType,
    scanDataFilePath,
    numProc,
    chunkSize,
    fileLimit,
    scanType="full",
    inputPathList=None,
    pathListFilePath=None,
    dataCoverageFilePath=None,
    dataTypeFilePath=None,
    failedFilePath=None,
    cachePath=None,
):
    """ Utility method to scan the data repository of the input content type and store type and coverage details.
    """
    try:
        #
        # configName = cfgOb.getDefaultSectionName()
        dP = DictionaryApiProviderWrapper(cfgOb, cachePath, useCache=True)
        dictApi = dP.getApiByName(contentType)
        ###
        categoryList = sorted(dictApi.getCategoryList())
        dictSchema = {catName: sorted(dictApi.getAttributeNameList(catName)) for catName in categoryList}
        attributeDataTypeD = OrderedDict()
        for catName in categoryList:
            aD = {}
            for atName in dictSchema[catName]:
                aD[atName] = dictApi.getTypeCode(catName, atName)
            attributeDataTypeD[catName] = aD
        ###
        #
        sr = ScanRepoUtil(cfgOb, attributeDataTypeD=attributeDataTypeD, numProc=numProc, chunkSize=chunkSize, fileLimit=fileLimit, workPath=cachePath)
        ok = sr.scanContentType(
            contentType, scanType=scanType, inputPathList=inputPathList, scanDataFilePath=scanDataFilePath, failedFilePath=failedFilePath, saveInputFileListPath=pathListFilePath
        )
        if dataTypeFilePath:
            ok = sr.evalScan(scanDataFilePath, dataTypeFilePath, evalType="data_type")
        if dataCoverageFilePath:
            ok = sr.evalScan(scanDataFilePath, dataCoverageFilePath, evalType="data_coverage")

        return ok
    except Exception as e:
        logger.exception("Failing with %s", str(e))


def main():
    parser = argparse.ArgumentParser()
    defaultConfigName = "site_info_configuration"
    #
    parser.add_argument("--scanType", default="full", help="Repository scan type (full|incr)")
    #
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--scan_chem_comp_ref", default=False, action="store_true", help="Scan Chemical Component reference definitions (public subset)")
    group.add_argument("--scan_chem_comp_core_ref", default=False, action="store_true", help="Scan Chemical Component Core reference definitions (public subset)")
    group.add_argument("--scan_bird_chem_comp_ref", default=False, action="store_true", help="Scan Bird Chemical Component reference definitions (public subset)")
    group.add_argument("--scan_bird_chem_comp_core_ref", default=False, action="store_true", help="Scan Bird Chemical Component Core reference definitions (public subset)")
    group.add_argument("--scan_bird_ref", default=False, action="store_true", help="Scan Bird reference definitions (public subset)")
    group.add_argument("--scan_bird_family_ref", default=False, action="store_true", help="Scan Bird Family reference definitions (public subset)")
    group.add_argument("--scan_entry_data", default=False, action="store_true", help="Scan PDB entry data (current released subset)")
    group.add_argument("--scan_ihm_dev", default=False, action="store_true", help="Scan PDBDEV I/HM entry data (current released subset)")
    #
    parser.add_argument("--config_path", default=None, help="Path to configuration options file")
    parser.add_argument("--config_name", default=defaultConfigName, help="Configuration section name")

    parser.add_argument("--input_file_list_path", default=None, help="Input file containing file paths to scan")
    parser.add_argument("--output_file_list_path", default=None, help="Output file containing file paths scanned")
    parser.add_argument("--fail_file_list_path", default=None, help="Output file containing file paths that fail scan")
    parser.add_argument("--scan_data_file_path", default=None, help="Output working file storing scan data (Pickle)")
    parser.add_argument("--coverage_file_path", default=None, help="Coverage map (JSON) output path")
    parser.add_argument("--type_map_file_path", default=None, help="Type map (JSON) output path")

    parser.add_argument("--num_proc", default=2, help="Number of processes to execute (default=2)")
    parser.add_argument("--chunk_size", default=10, help="Number of files loaded per process")
    parser.add_argument("--file_limit", default=None, help="Load file limit for testing")
    parser.add_argument("--debug", default=False, action="store_true", help="Turn on verbose logging")
    parser.add_argument("--mock", default=False, action="store_true", help="Use MOCK repository configuration for testing")
    parser.add_argument("--cache_path", default=None, help="Cache path and working direcory for temporary files")
    args = parser.parse_args()
    #
    debugFlag = args.debug
    if debugFlag:
        logger.setLevel(logging.DEBUG)
    # ----------------------- - ----------------------- - ----------------------- - ----------------------- - ----------------------- -
    #                                       Configuration Details
    configPath = args.config_path
    configName = args.config_name
    if not configPath:
        configPath = os.getenv("DBLOAD_CONFIG_PATH", None)
    try:
        if os.access(configPath, os.R_OK):
            os.environ["DBLOAD_CONFIG_PATH"] = configPath
            logger.info("Using configuation path %s (%s)", configPath, configName)
        else:
            logger.error("Missing or access issue with config file %r", configPath)
            exit(1)
        mockTopPath = os.path.join(TOPDIR, "rcsb", "mock-data") if args.mock else None
        cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=defaultConfigName, mockTopPath=mockTopPath)
        if configName != defaultConfigName:
            cfgOb.replaceSectionName(defaultConfigName, configName)
    except Exception as e:
        logger.error("Missing or access issue with config file %r with %s", configPath, str(e))
        exit(1)

    #
    try:
        numProc = int(args.num_proc)
        chunkSize = int(args.chunk_size)
        fileLimit = int(args.file_limit) if args.file_limit else None
        #
        failedFilePath = args.fail_file_list_path

        scanType = args.scanType
        #
        inputFileListPath = args.input_file_list_path
        outputFileListPath = args.output_file_list_path
        scanDataFilePath = args.scan_data_file_path
        dataCoverageFilePath = args.coverage_file_path
        dataTypeFilePath = args.type_map_file_path
        cachePath = args.cache_path if args.cache_path else "."
    except Exception as e:
        logger.exception("Argument processing problem %s", str(e))
        parser.print_help(sys.stderr)
        exit(1)
    # ----------------------- - ----------------------- - ----------------------- - ----------------------- - ----------------------- -
    #
    # Read any input path lists -
    #
    inputPathList = None
    if inputFileListPath:
        mu = MarshalUtil(workPath=cachePath)
        inputPathList = mu.doImport(inputFileListPath, fmt="list")
    #
    ##

    if args.scan_chem_comp_ref:
        contentType = "chem_comp_core"

    elif args.scan_chem_comp_core_ref:
        contentType = "chem_comp_core"

    elif args.scan_bird_chem_comp_ref:
        contentType = "bird_chem_comp_core"

    elif args.scan_bird_chem_comp_core_ref:
        contentType = "bird_chem_comp_core"

    elif args.scan_bird_ref:
        contentType = "bird"

    elif args.scan_bird_family_ref:
        contentType = "bird_family"

    elif args.scan_entry_data:
        contentType = "pdbx"

    elif args.scan_ihm_dev:
        contentType = "ihm_dev"

    ok = scanRepo(
        cfgOb,
        contentType,
        scanDataFilePath,
        numProc,
        chunkSize,
        fileLimit,
        scanType=scanType,
        inputPathList=inputPathList,
        pathListFilePath=outputFileListPath,
        dataCoverageFilePath=dataCoverageFilePath,
        dataTypeFilePath=dataTypeFilePath,
        failedFilePath=failedFilePath,
        cachePath=cachePath,
    )

    logger.info("Operation completed with status %r", ok)


if __name__ == "__main__":
    main()



##
# File:    DictMethodAssemblyHelper.py
# Author:  J. Westbrook
# Date:    16-Jul-2019
# Version: 0.001 Initial version
#
##
"""
Helper class implementing external assembly-level methods  supporting the RCSB dictionary extension.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import re
from collections import Counter

from mmcif.api.DataCategory import DataCategory

logger = logging.getLogger(__name__)


def cmpElements(lhs, rhs):
    return 0 if (lhs[-1].isdigit() or lhs[-1] in ["R", "S"]) and rhs[0].isdigit() else -1


class DictMethodAssemblyHelper(object):
    """Helper class implementing external assembly-level methods  supporting the RCSB dictionary extension."""

    def __init__(self, **kwargs):
        """
        Args:
            **kwargs: (dict)  Placeholder for future key-value arguments

        """
        #
        self._raiseExceptions = kwargs.get("raiseExceptions", False)
        #
        rP = kwargs.get("resourceProvider")
        self.__commonU = rP.getResource("DictMethodCommonUtils instance") if rP else None
        self.__dApi = rP.getResource("Dictionary API instance (pdbx_core)") if rP else None
        #
        logger.debug("Dictionary method helper init")

    def echo(self, msg):
        logger.info(msg)

    def addAssemblyInfo(self, dataContainer, catName, **kwargs):
        """Build rcsb_assembly_info category.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise

        """
        logger.debug("Starting catName %s kwargs %r", catName, kwargs)
        try:
            if not (dataContainer.exists("entry") and dataContainer.exists("pdbx_struct_assembly")):
                return False
            logger.debug("%s beginning for %s", dataContainer.getName(), catName)
            # Create the new target category rcsb_assembly_info
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=self.__dApi.getAttributeNameList(catName)))
            #
            #
            logger.debug("%s beginning for %s", dataContainer.getName(), catName)
            #
            # Get assembly comp details -
            #
            rD = self.__getAssemblyComposition(dataContainer)
            #
            cObj = dataContainer.getObj(catName)

            tObj = dataContainer.getObj("entry")
            entryId = tObj.getValue("id", 0)
            #
            tObj = dataContainer.getObj("pdbx_struct_assembly")
            assemblyIdL = tObj.getAttributeValueList("id")
            #
            #
            for ii, assemblyId in enumerate(assemblyIdL):
                if assemblyId not in rD["assemblyHeavyAtomCountByTypeD"]:
                    continue
                if assemblyId not in rD["assemblyHeavyAtomCountD"]:
                    continue
                dD = rD["assemblyHeavyAtomCountByTypeD"][assemblyId]
                #
                cObj.setValue(entryId, "entry_id", ii)
                cObj.setValue(assemblyId, "assembly_id", ii)
                #

                num = dD["polymer"] if "polymer" in dD else 0
                cObj.setValue(num, "polymer_atom_count", ii)

                num = dD["non-polymer"] if "non-polymer" in dD else 0
                cObj.setValue(num, "nonpolymer_atom_count", ii)

                num = dD["water"] if "water" in dD else 0
                cObj.setValue(num, "solvent_atom_count", ii)

                num = dD["branched"] if "branched" in dD else 0
                cObj.setValue(num, "branched_atom_count", ii)

                num = rD["assemblyHeavyAtomCountD"][assemblyId]
                cObj.setValue(num, "atom_count", ii)
                #
                num = rD["assemblyHydrogenAtomCountD"][assemblyId]
                cObj.setValue(num, "hydrogen_atom_count", ii)
                #
                num1 = rD["assemblyModeledMonomerCountD"][assemblyId]
                num2 = rD["assemblyUnmodeledMonomerCountD"][assemblyId]
                cObj.setValue(num1, "modeled_polymer_monomer_count", ii)
                cObj.setValue(num2, "unmodeled_polymer_monomer_count", ii)
                cObj.setValue(num1 + num2, "polymer_monomer_count", ii)
                #
                dD = rD["assemblyPolymerClassD"][assemblyId]
                cObj.setValue(dD["polymerCompClass"], "polymer_composition", ii)
                cObj.setValue(dD["subsetCompClass"], "selected_polymer_entity_types", ii)
                cObj.setValue(dD["naCompClass"], "na_polymer_entity_types", ii)
                #
                dD = rD["assemblyInstanceCountByTypeD"][assemblyId]
                num = dD["polymer"] if "polymer" in dD else 0
                cObj.setValue(num, "polymer_entity_instance_count", ii)
                #
                num = dD["non-polymer"] if "non-polymer" in dD else 0
                cObj.setValue(num, "nonpolymer_entity_instance_count", ii)
                #
                num = dD["branched"] if "branched" in dD else 0
                cObj.setValue(num, "branched_entity_instance_count", ii)
                #
                num = dD["water"] if "water" in dD else 0
                cObj.setValue(num, "solvent_entity_instance_count", ii)
                #
                dD = rD["assemblyInstanceCountByPolymerTypeD"][assemblyId]
                num = dD["Protein"] if "Protein" in dD else 0
                cObj.setValue(num, "polymer_entity_instance_count_protein", ii)
                num1 = dD["DNA"] if "DNA" in dD else 0
                cObj.setValue(num1, "polymer_entity_instance_count_DNA", ii)
                num2 = dD["RNA"] if "RNA" in dD else 0
                cObj.setValue(num2, "polymer_entity_instance_count_RNA", ii)
                cObj.setValue(num1 + num2, "polymer_entity_instance_count_nucleic_acid", ii)
                num = dD["NA-hybrid"] if "NA-hybrid" in dD else 0
                cObj.setValue(num, "polymer_entity_instance_count_nucleic_acid_hybrid", ii)
                #
                dD = rD["assemblyEntityCountByPolymerTypeD"][assemblyId]
                num = dD["Protein"] if "Protein" in dD else 0
                cObj.setValue(num, "polymer_entity_count_protein", ii)
                num1 = dD["DNA"] if "DNA" in dD else 0
                cObj.setValue(num1, "polymer_entity_count_DNA", ii)
                num2 = dD["RNA"] if "RNA" in dD else 0
                cObj.setValue(num2, "polymer_entity_count_RNA", ii)
                cObj.setValue(num1 + num2, "polymer_entity_count_nucleic_acid", ii)
                num = dD["NA-hybrid"] if "NA-hybrid" in dD else 0
                cObj.setValue(num, "polymer_entity_count_nucleic_acid_hybrid", ii)
                #
                dD = rD["assemblyEntityCountByTypeD"][assemblyId]
                num = dD["polymer"] if "polymer" in dD else 0
                cObj.setValue(num, "polymer_entity_count", ii)
                #
                num = dD["non-polymer"] if "non-polymer" in dD else 0
                cObj.setValue(num, "nonpolymer_entity_count", ii)
                #
                num = dD["branched"] if "branched" in dD else 0
                cObj.setValue(num, "branched_entity_count", ii)
                #
                num = dD["water"] if "water" in dD else 0
                cObj.setValue(num, "solvent_entity_count", ii)
            #
            return
        except Exception as e:
            logger.exception("For %s failing with %s", catName, str(e))
        return False

    def buildContainerAssemblyIds(self, dataContainer, catName, **kwargs):
        """Build category rcsb_assembly_container_identifiers.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise

        For example,

        loop_
        _rcsb_assembly_container_identifiers.entry_id
        _rcsb_assembly_container_identifiers.assembly_id
        ...


        """
        logger.debug("Starting catName %s kwargs %r", catName, kwargs)
        try:
            if not (dataContainer.exists("entry") and dataContainer.exists("pdbx_struct_assembly")):
                return False
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=self.__dApi.getAttributeNameList(catName)))
            #
            cObj = dataContainer.getObj(catName)

            tObj = dataContainer.getObj("entry")
            entryId = tObj.getValue("id", 0)
            cObj.setValue(entryId, "entry_id", 0)
            #
            tObj = dataContainer.getObj("pdbx_struct_assembly")
            assemblyIdL = tObj.getAttributeValueList("id")
            for ii, assemblyId in enumerate(assemblyIdL):
                cObj.setValue(entryId, "entry_id", ii)
                cObj.setValue(assemblyId, "assembly_id", ii)
                cObj.setValue(entryId + "-" + assemblyId, "rcsb_id", ii)

            #
            return True
        except Exception as e:
            logger.exception("For %s failing with %s", catName, str(e))
        return False

    def addDepositedAssembly(self, dataContainer, catName, **kwargs):
        """Add the deposited coordinates as an additional separate assembly labeled as 'deposited'
        to categories, pdbx_struct_assembly and pdb_struct_assembly_gen.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise

        """
        logger.debug("Starting catName %s kwargs %r", catName, kwargs)
        try:
            if not dataContainer.exists("struct_asym"):
                return False
            if not dataContainer.exists("pdbx_struct_assembly"):
                dataContainer.append(
                    DataCategory(
                        "pdbx_struct_assembly",
                        attributeNameList=["id", "details", "method_details", "oligomeric_details", "oligomeric_count", "rcsb_details", "rcsb_candidate_assembly"],
                    )
                )
            if not dataContainer.exists("pdbx_struct_assembly_gen"):
                dataContainer.append(DataCategory("pdbx_struct_assembly_gen", attributeNameList=["assembly_id", "oper_expression", "asym_id_list", "ordinal"]))

            if not dataContainer.exists("pdbx_struct_oper_list"):
                row = [
                    "1",
                    "identity operation",
                    "1_555",
                    "x, y, z",
                    "1.0000000000",
                    "0.0000000000",
                    "0.0000000000",
                    "0.0000000000",
                    "0.0000000000",
                    "1.0000000000",
                    "0.0000000000",
                    "0.0000000000",
                    "0.0000000000",
                    "0.0000000000",
                    "1.0000000000",
                    "0.0000000000",
                ]
                atList = [
                    "id",
                    "type",
                    "name",
                    "symmetry_operation",
                    "matrix[1][1]",
                    "matrix[1][2]",
                    "matrix[1][3]",
                    "vector[1]",
                    "matrix[2][1]",
                    "matrix[2][2]",
                    "matrix[2][3]",
                    "vector[2]",
                    "matrix[3][1]",
                    "matrix[3][2]",
                    "matrix[3][3]",
                    "vector[3]",
                ]
                dataContainer.append(DataCategory("pdbx_struct_oper_list", attributeNameList=atList, rowList=[row]))

            #
            logger.debug("Add deposited assembly for %s", dataContainer.getName())
            cObj = dataContainer.getObj("struct_asym")
            asymIdL = cObj.getAttributeValueList("id")
            logger.debug("AsymIdL %r", asymIdL)
            #
            # Ordinal is added by subsequent attribure-level method.
            tObj = dataContainer.getObj("pdbx_struct_assembly_gen")
            rowIdx = tObj.getRowCount()
            tObj.setValue("deposited", "assembly_id", rowIdx)
            tObj.setValue("1", "oper_expression", rowIdx)
            tObj.setValue(",".join(asymIdL), "asym_id_list", rowIdx)
            #
            tObj = dataContainer.getObj("pdbx_struct_assembly")
            rowIdx = tObj.getRowCount()
            tObj.setValue("deposited", "id", rowIdx)
            tObj.setValue("deposited_coordinates", "details", rowIdx)
            #
            for atName in ["oligomeric_details", "method_details", "oligomeric_count"]:
                if tObj.hasAttribute(atName):
                    tObj.setValue("?", atName, rowIdx)
            #
            #
            #
            logger.debug("Full row is %r", tObj.getRow(rowIdx))
            #
            return True
        except Exception as e:
            logger.exception("For %s failing with %s", catName, str(e))
        return False

    def filterAssemblyDetails(self, dataContainer, catName, **kwargs):
        """Filter _pdbx_struct_assembly.details -> _pdbx_struct_assembly.rcsb_details
            with a more limited vocabulary -


        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise

        For example, mapping to the following limited enumeration,

                'author_and_software_defined_assembly'
                'author_defined_assembly'
                'software_defined_assembly'

        """
        logger.debug("Starting catName %s kwargs %r", catName, kwargs)
        mD = {
            "author_and_software_defined_assembly": "author_and_software_defined_assembly",
            "author_defined_assembly": "author_defined_assembly",
            "complete icosahedral assembly": "author_and_software_defined_assembly",
            "complete point assembly": "author_and_software_defined_assembly",
            "crystal asymmetric unit": "software_defined_assembly",
            "crystal asymmetric unit, crystal frame": "software_defined_assembly",
            "details": "software_defined_assembly",
            "helical asymmetric unit": "software_defined_assembly",
            "helical asymmetric unit, std helical frame": "software_defined_assembly",
            "icosahedral 23 hexamer": "software_defined_assembly",
            "icosahedral asymmetric unit": "software_defined_assembly",
            "icosahedral asymmetric unit, std point frame": "software_defined_assembly",
            "icosahedral pentamer": "software_defined_assembly",
            "pentasymmetron capsid unit": "software_defined_assembly",
            "point asymmetric unit": "software_defined_assembly",
            "point asymmetric unit, std point frame": "software_defined_assembly",
            "representative helical assembly": "author_and_software_defined_assembly",
            "software_defined_assembly": "software_defined_assembly",
            "trisymmetron capsid unit": "software_defined_assembly",
            "deposited_coordinates": "software_defined_assembly",
        }
        #
        try:
            if not dataContainer.exists("pdbx_struct_assembly"):
                return False

            logger.debug("Filter assembly details for %s", dataContainer.getName())
            tObj = dataContainer.getObj("pdbx_struct_assembly")
            atName = "rcsb_details"
            if not tObj.hasAttribute(atName):
                tObj.appendAttribute(atName)
            #
            for iRow in range(tObj.getRowCount()):
                details = tObj.getValue("details", iRow)
                if details in mD:
                    tObj.setValue(mD[details], "rcsb_details", iRow)
                else:
                    tObj.setValue("software_defined_assembly", "rcsb_details", iRow)
                # logger.debug("Full row is %r", tObj.getRow(iRow))
            return True
        except Exception as e:
            logger.exception("For %s %s failing with %s", catName, atName, str(e))
        return False

    def assignAssemblyCandidates(self, dataContainer, catName, **kwargs):
        """Flag candidate biological assemblies as 'author_defined_assembly' ad author_and_software_defined_assembly'

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise

        """
        logger.debug("Starting catName %s kwargs %r", catName, kwargs)
        mD = {
            "author_and_software_defined_assembly": "author_and_software_defined_assembly",
            "author_defined_assembly": "author_defined_assembly",
            "complete icosahedral assembly": "author_and_software_defined_assembly",
            "complete point assembly": "author_and_software_defined_assembly",
            "crystal asymmetric unit": "software_defined_assembly",
            "crystal asymmetric unit, crystal frame": "software_defined_assembly",
            "details": "software_defined_assembly",
            "helical asymmetric unit": "software_defined_assembly",
            "helical asymmetric unit, std helical frame": "software_defined_assembly",
            "icosahedral 23 hexamer": "software_defined_assembly",
            "icosahedral asymmetric unit": "software_defined_assembly",
            "icosahedral asymmetric unit, std point frame": "software_defined_assembly",
            "icosahedral pentamer": "software_defined_assembly",
            "pentasymmetron capsid unit": "software_defined_assembly",
            "point asymmetric unit": "software_defined_assembly",
            "point asymmetric unit, std point frame": "software_defined_assembly",
            "representative helical assembly": "author_and_software_defined_assembly",
            "software_defined_assembly": "software_defined_assembly",
            "trisymmetron capsid unit": "software_defined_assembly",
            "deposited_coordinates": "software_defined_assembly",
        }
        #
        eD = {
            k: True
            for k in [
                "crystal asymmetric unit",
                "crystal asymmetric unit, crystal frame",
                "helical asymmetric unit",
                "helical asymmetric unit, std helical frame",
                "icosahedral 23 hexamer",
                "icosahedral asymmetric unit",
                "icosahedral asymmetric unit, std point frame",
                "icosahedral pentamer",
                "pentasymmetron capsid unit",
                "point asymmetric unit",
                "point asymmetric unit, std point frame",
                "trisymmetron capsid unit",
                "deposited_coordinates",
                "details",
            ]
        }
        try:
            if not dataContainer.exists("pdbx_struct_assembly"):
                return False
            atName = "rcsb_candidate_assembly"
            tObj = dataContainer.getObj("pdbx_struct_assembly")
            if not tObj.hasAttribute(atName):
                tObj.appendAttribute(atName)
            #
            for iRow in range(tObj.getRowCount()):
                details = tObj.getValue("details", iRow)
                if details in mD and details not in eD:
                    tObj.setValue("Y", "rcsb_candidate_assembly", iRow)
                else:
                    tObj.setValue("N", "rcsb_candidate_assembly", iRow)
                # logger.debug("Full row is %r", tObj.getRow(iRow))

            #
            return True
        except Exception as e:
            logger.exception("For %s %s failing with %s", catName, atName, str(e))
        return False

    def filterAssemblyCandidates(self, dataContainer, catName, **kwargs):
        """Filter assemblies to only candidates and deposited cases


        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise


        """
        logger.debug("Starting catName %s kwargs %r", catName, kwargs)
        try:
            if not dataContainer.exists("pdbx_struct_assembly"):
                return False

            logger.debug("Filter candidate assemblyfor %s", dataContainer.getName())
            tObj = dataContainer.getObj("pdbx_struct_assembly")
            #
            indexList = []
            for iRow in range(tObj.getRowCount()):
                isCandidate = tObj.getValue("rcsb_candidate_assembly", iRow) == "Y"
                isDeposited = tObj.getValue("id", iRow) == "deposited"

                if not (isCandidate or isDeposited):
                    indexList.append(iRow)
            tObj.removeRows(indexList)
            #
            # ---
            numAssemblies = tObj.getRowCount()
            logger.debug("Assembly count is %d", numAssemblies)
            if dataContainer.exists("rcsb_entry_info"):
                eiObj = dataContainer.getObj("rcsb_entry_info")
                eiObj.setValue(numAssemblies, "assembly_count", 0)
            #
            return True
        except Exception as e:
            logger.exception("For %s failing with %s", catName, str(e))
        return False

    def __expandOperatorList(self, operExpression):
        """
        Operation expressions may have the forms:

                (1)        the single operation 1
                (1,2,5)    the operations 1, 2, 5
                (1-4)      the operations 1,2,3 and 4
                (1,2)(3,4) the combinations of operations
                           3 and 4 followed by 1 and 2 (i.e.
                           the cartesian product of parenthetical
                           groups applied from right to left)
        """

        rL = []
        opCount = 1
        try:
            if operExpression.find("(") < 0:
                opL = [operExpression]
            else:
                opL = [tV.strip().strip("(").rstrip(")") for tV in re.findall(r"\(.*?\)", operExpression)]
            #
            for op in opL:
                teL = []
                tL = op.split(",")
                for tV in tL:
                    trngL = tV.split("-")
                    if len(trngL) == 2:
                        rngL = [str(r) for r in range(int(trngL[0]), int(trngL[1]) + 1)]
                    else:
                        rngL = trngL
                    teL.extend(rngL)
                rL.append(teL)
                opCount *= len(teL)

        except Exception as e:
            logger.exception("Failing parsing %r with %s", operExpression, str(e))
        #
        if not rL:
            opCount = 0
        return opCount, rL

    def __getAssemblyComposition(self, dataContainer):
        """Return assembly composition by entity and instance type counts.

        Example -
            loop_
            _pdbx_struct_assembly.id
            _pdbx_struct_assembly.details
            _pdbx_struct_assembly.method_details
            _pdbx_struct_assembly.oligomeric_details
            _pdbx_struct_assembly.oligomeric_count
            1 'complete icosahedral assembly'                ? 180-meric      180
            2 'icosahedral asymmetric unit'                  ? trimeric       3
            3 'icosahedral pentamer'                         ? pentadecameric 15
            4 'icosahedral 23 hexamer'                       ? octadecameric  18
            5 'icosahedral asymmetric unit, std point frame' ? trimeric       3
            #
            loop_
            _pdbx_struct_assembly_gen.assembly_id
            _pdbx_struct_assembly_gen.oper_expression
            _pdbx_struct_assembly_gen.asym_id_list
            1 '(1-60)'           A,B,C
            2 1                  A,B,C
            3 '(1-5)'            A,B,C
            4 '(1,2,6,10,23,24)' A,B,C
            5 P                  A,B,C
            #
        """
        #
        instanceTypeD = self.__commonU.getInstanceTypes(dataContainer)
        instancePolymerTypeD = self.__commonU.getInstancePolymerTypes(dataContainer)
        instEntityD = self.__commonU.getInstanceEntityMap(dataContainer)
        #
        epTypeD = self.__commonU.getEntityPolymerTypes(dataContainer)
        eTypeD = self.__commonU.getEntityTypes(dataContainer)
        epTypeFilteredD = self.__commonU.getPolymerEntityFilteredTypes(dataContainer)
        # JDW
        instHeavyAtomCount = self.__commonU.getInstanceHeavyAtomCounts(dataContainer, modelId="1")
        instHydrogenAtomCount = self.__commonU.getInstanceHydrogenAtomCounts(dataContainer, modelId="1")
        #
        instModeledMonomerCount = self.__commonU.getInstanceModeledMonomerCounts(dataContainer, modelId="1")
        instUnmodeledMonomerCount = self.__commonU.getInstanceUnModeledMonomerCounts(dataContainer, modelId="1")
        # -------------------------
        assemblyInstanceCountByTypeD = {}
        assemblyHeavyAtomCountByTypeD = {}
        assemblyHeavyAtomCountD = {}
        assemblyHydrogenAtomCountD = {}
        assemblyModeledMonomerCountD = {}
        assemblyUnmodeledMonomerCountD = {}
        # Pre-generation (source instances)
        assemblyInstanceD = {}
        # Post-generation (gerated instances)
        assemblyInstanceGenD = {}
        assemblyInstanceCountByPolymerTypeD = {}
        assemblyPolymerInstanceCountD = {}
        assemblyPolymerClassD = {}
        #
        assemblyEntityCountByPolymerTypeD = {}
        assemblyEntityCountByTypeD = {}
        # --------------
        #
        try:
            if dataContainer.exists("pdbx_struct_assembly_gen"):
                tObj = dataContainer.getObj("pdbx_struct_assembly_gen")
                for ii in range(tObj.getRowCount()):
                    assemblyId = tObj.getValue("assembly_id", ii)
                    # Initialize instances count
                    if assemblyId not in assemblyInstanceCountByTypeD:
                        assemblyInstanceCountByTypeD[assemblyId] = {eType: 0 for eType in ["polymer", "non-polymer", "branched", "macrolide", "water"]}
                    if assemblyId not in assemblyHeavyAtomCountByTypeD:
                        assemblyHeavyAtomCountByTypeD[assemblyId] = {eType: 0 for eType in ["polymer", "non-polymer", "branched", "macrolide", "water"]}
                    if assemblyId not in assemblyModeledMonomerCountD:
                        assemblyModeledMonomerCountD[assemblyId] = 0
                    if assemblyId not in assemblyUnmodeledMonomerCountD:
                        assemblyUnmodeledMonomerCountD[assemblyId] = 0
                    if assemblyId not in assemblyHeavyAtomCountD:
                        assemblyHeavyAtomCountD[assemblyId] = 0
                    if assemblyId not in assemblyHydrogenAtomCountD:
                        assemblyHydrogenAtomCountD[assemblyId] = 0
                    #
                    opExpression = tObj.getValue("oper_expression", ii)
                    opCount, opL = self.__expandOperatorList(opExpression)
                    tS = tObj.getValue("asym_id_list", ii)
                    asymIdList = [t.strip() for t in tS.strip().split(",")]
                    assemblyInstanceD.setdefault(assemblyId, []).extend(asymIdList)
                    assemblyInstanceGenD.setdefault(assemblyId, []).extend(asymIdList * opCount)
                    #
                    logger.debug("%s assembly %r opExpression %r opCount %d opL %r", dataContainer.getName(), assemblyId, opExpression, opCount, opL)
                    logger.debug("%s assembly %r length asymIdList %r", dataContainer.getName(), assemblyId, len(asymIdList))
                    #
                    for eType in ["polymer", "non-polymer", "branched", "macrolide", "water"]:
                        iList = [asymId for asymId in asymIdList if asymId in instanceTypeD and instanceTypeD[asymId] == eType]
                        assemblyInstanceCountByTypeD[assemblyId][eType] += len(iList) * opCount
                        #
                        atCountList = [
                            instHeavyAtomCount[asymId] for asymId in asymIdList if asymId in instanceTypeD and instanceTypeD[asymId] == eType and asymId in instHeavyAtomCount
                        ]
                        assemblyHeavyAtomCountByTypeD[assemblyId][eType] += sum(atCountList) * opCount
                        assemblyHeavyAtomCountD[assemblyId] += sum(atCountList) * opCount
                        #
                        hAtCountList = [
                            instHydrogenAtomCount[asymId] for asymId in asymIdList if asymId in instanceTypeD and instanceTypeD[asymId] == eType and asymId in instHydrogenAtomCount
                        ]
                        assemblyHydrogenAtomCountD[assemblyId] += sum(hAtCountList) * opCount
                    #
                    modeledMonomerCountList = [
                        instModeledMonomerCount[asymId]
                        for asymId in asymIdList
                        if asymId in instanceTypeD and instanceTypeD[asymId] == "polymer" and asymId in instModeledMonomerCount
                    ]
                    assemblyModeledMonomerCountD[assemblyId] += sum(modeledMonomerCountList) * opCount
                    #
                    unmodeledMonomerCountList = [
                        instUnmodeledMonomerCount[asymId]
                        for asymId in asymIdList
                        if asymId in instanceTypeD and instanceTypeD[asymId] == "polymer" and asymId in instUnmodeledMonomerCount
                    ]
                    assemblyUnmodeledMonomerCountD[assemblyId] += sum(unmodeledMonomerCountList) * opCount

                #
                assemblyInstanceCountByPolymerTypeD = {}
                assemblyPolymerInstanceCountD = {}
                assemblyPolymerClassD = {}
                #
                assemblyEntityCountByPolymerTypeD = {}
                assemblyEntityCountByTypeD = {}
                #
                # Using the generated list of instance assembly components ...
                for assemblyId, asymIdList in assemblyInstanceGenD.items():
                    # ------
                    #  Instance polymer composition
                    pInstTypeList = [instancePolymerTypeD[asymId] for asymId in asymIdList if asymId in instancePolymerTypeD]
                    pInstTypeD = Counter(pInstTypeList)
                    assemblyInstanceCountByPolymerTypeD[assemblyId] = {pType: 0 for pType in ["Protein", "DNA", "RNA", "NA-hybrid", "Other"]}
                    assemblyInstanceCountByPolymerTypeD[assemblyId] = {pType: pInstTypeD[pType] for pType in ["Protein", "DNA", "RNA", "NA-hybrid", "Other"] if pType in pInstTypeD}
                    assemblyPolymerInstanceCountD[assemblyId] = len(pInstTypeList)
                    #
                    logger.debug("%s assemblyId %r pInstTypeD %r", dataContainer.getName(), assemblyId, pInstTypeD.items())

                    # -------------
                    # Entity and polymer entity composition
                    #
                    entityIdList = list(set([instEntityD[asymId] for asymId in asymIdList if asymId in instEntityD]))
                    pTypeL = [epTypeD[entityId] for entityId in entityIdList if entityId in epTypeD]
                    #
                    polymerCompClass, subsetCompClass, naCompClass, _ = self.__commonU.getPolymerComposition(pTypeL)
                    assemblyPolymerClassD[assemblyId] = {"polymerCompClass": polymerCompClass, "subsetCompClass": subsetCompClass, "naCompClass": naCompClass}
                    #
                    logger.debug(
                        "%s assemblyId %s polymerCompClass %r subsetCompClass %r naCompClass %r pTypeL %r",
                        dataContainer.getName(),
                        assemblyId,
                        polymerCompClass,
                        subsetCompClass,
                        naCompClass,
                        pTypeL,
                    )
                    pTypeFilteredL = [epTypeFilteredD[entityId] for entityId in entityIdList if entityId in epTypeFilteredD]
                    #
                    pEntityTypeD = Counter(pTypeFilteredL)
                    assemblyEntityCountByPolymerTypeD[assemblyId] = {pType: 0 for pType in ["Protein", "DNA", "RNA", "NA-hybrid", "Other"]}
                    assemblyEntityCountByPolymerTypeD[assemblyId] = {
                        pType: pEntityTypeD[pType] for pType in ["Protein", "DNA", "RNA", "NA-hybrid", "Other"] if pType in pEntityTypeD
                    }
                    #
                    eTypeL = [eTypeD[entityId] for entityId in entityIdList if entityId in eTypeD]
                    entityTypeD = Counter(eTypeL)
                    assemblyEntityCountByTypeD[assemblyId] = {eType: 0 for eType in ["polymer", "non-polymer", "branched", "macrolide", "water"]}
                    assemblyEntityCountByTypeD[assemblyId] = {
                        eType: entityTypeD[eType] for eType in ["polymer", "non-polymer", "branched", "macrolide", "water"] if eType in entityTypeD
                    }
                    #
                    # ---------------
                    #
            #
            logger.debug("%s assemblyInstanceCountByTypeD %r", dataContainer.getName(), assemblyInstanceCountByTypeD.items())
            logger.debug("%s assemblyHeavyAtomCountByTypeD %r", dataContainer.getName(), assemblyHeavyAtomCountByTypeD.items())
            logger.debug("%s assemblyHeavyAtomCountD %r", dataContainer.getName(), assemblyHeavyAtomCountD.items())
            logger.debug("%s assemblyHydrogenAtomCountD %r", dataContainer.getName(), assemblyHydrogenAtomCountD.items())
            logger.debug("%s assemblyModeledMonomerCountD %r", dataContainer.getName(), assemblyModeledMonomerCountD.items())
            logger.debug("%s assemblyUnmodeledMonomerCountD %r", dataContainer.getName(), assemblyUnmodeledMonomerCountD.items())
            logger.debug("%s assemblyPolymerClassD %r", dataContainer.getName(), assemblyPolymerClassD.items())
            logger.debug("%s assemblyPolymerInstanceCountD %r", dataContainer.getName(), assemblyPolymerInstanceCountD.items())
            logger.debug("%s assemblyInstanceCountByPolymerTypeD %r", dataContainer.getName(), assemblyInstanceCountByPolymerTypeD.items())
            logger.debug("%s assemblyEntityCountByPolymerTypeD %r", dataContainer.getName(), assemblyEntityCountByPolymerTypeD.items())
            logger.debug("%s assemblyEntityCountByTypeD %r", dataContainer.getName(), assemblyEntityCountByTypeD.items())
            #
            rD = {
                "assemblyInstanceCountByTypeD": assemblyInstanceCountByTypeD,
                "assemblyHeavyAtomCountByTypeD": assemblyHeavyAtomCountByTypeD,
                "assemblyHeavyAtomCountD": assemblyHeavyAtomCountD,
                "assemblyHydrogenAtomCountD": assemblyHydrogenAtomCountD,
                "assemblyModeledMonomerCountD": assemblyModeledMonomerCountD,
                "assemblyUnmodeledMonomerCountD": assemblyUnmodeledMonomerCountD,
                "assemblyInstanceCountByPolymerTypeD": assemblyInstanceCountByPolymerTypeD,
                "assemblyPolymerInstanceCountD": assemblyPolymerInstanceCountD,
                "assemblyPolymerClassD": assemblyPolymerClassD,
                "assemblyEntityCountByPolymerTypeD": assemblyEntityCountByPolymerTypeD,
                "assemblyEntityCountByTypeD": assemblyEntityCountByTypeD,
            }
        except Exception as e:
            logger.exception("Failing %s with %s", dataContainer.getName(), str(e))
        return rD



##
# File:    DictMethodEntryHelper.py (DictMethodRunnerHelper.py)
# Author:  J. Westbrook
# Date:    18-Aug-2018
# Version: 0.001 Initial version
#
#
# Updates:
#  4-Sep-2018 jdw add methods to construct entry and entity identier categories.
# 10-Sep-2018 jdw add method for citation author aggregation
# 22-Sep-2018 jdw add method assignAssemblyCandidates()
# 27-Oct-2018 jdw add method consolidateAccessionDetails()
# 30-Oct-2018 jdw add category methods addChemCompRelated(), addChemCompInfo(),
#                 addChemCompDescriptor()
# 10-Nov-2018 jdw add addChemCompSynonyms(), addChemCompTargets(), filterBlockByMethod()
# 12-Nov-2018 jdw add InChIKey matching in addChemCompRelated()
# 15-Nov-2018 jdw add handling for antibody misrepresentation of multisource organisms
# 28-Nov-2018 jdw relax constraints on the production of rcsb_entry_info
#  1-Dec-2018 jdw add ncbi source and host organism info
# 11-Dec-2018 jdw add addStructRefSeqEntityIds and buildEntityPolySeq
# 10-Jan-2019 jdw better handle initialization in filterBlockByMethod()
# 11-Jan-2019 jdw revise classification in assignAssemblyCandidates()
# 16-Feb-2019 jdw add buildContainerEntityInstanceIds()
# 19-Feb-2019 jdw add internal method __addPdbxValidateAsymIds() to add cardinal identifiers to
#                 pdbx_validate_* categories
# 28-Feb-2019 jdw change criteria for adding rcsb_chem_comp_container_identiers to work with ion definitions
# 11-Mar-2019 jdw replace taxonomy file handling with calls to TaxonomyUtils()
# 11-Mar-2019 jdw add EC lineage using EnzymeDatabaseUtils()
# 17-Mar-2019 jdw add support for entity subcategory rcsb_macromolecular_names_combined
# 23-Mar-2019 jdw change criteria chem_comp collection criteria to _chem_comp.pdbx_release_status
# 25-Mar-2019 jdw remap merged taxons and adjust exception handling for taxonomy lineage generation
#  7-Apr-2019 jdw add CathClassificationUtils and CathClassificationUtils and sequence difference type counts
# 25-Apr-2019 jdw For source and host organism add ncbi_parent_scientific_name
#                 add rcsb_entry_info.deposited_modeled_polymer_monomer_count and
#                     rcsb_entry_info.deposited_unmodeled_polymer_monomer_count,
#  1-May-2019 jdw add support for _rcsb_entry_info.deposited_polymer_monomer_count,
#                   _rcsb_entry_info.polymer_entity_count_protein,
#                   _rcsb_entry_info.polymer_entity_count_nucleic_acid,
#                   _rcsb_entry_info.polymer_entity_count_nucleic_acid_hybrid,
#                   _rcsb_entry_info.polymer_entity_count_DNA,
#                   _rcsb_entry_info.polymer_entity_count_RNA,
#                   _rcsb_entry_info.nonpolymer_ligand_entity_count
#                   _rcsb_entry_info.selected_polymer_entity_types
#                   _rcsb_entry_info.polymer_entity_taxonomy_count
#                   _rcsb_entry_info.assembly_count
#                    add categories rcsb_entity_instance_domain_scop and rcsb_entity_instance_domain_cath
#  4-May-2019 jdw extend content in categories rcsb_entity_instance_domain_scop and rcsb_entity_instance_domain_cath
# 13-May-2019 jdw add rcsb_entry_info.deposited_polymer_entity_instance_count and deposited_nonpolymer_entity_instance_count
#                 add entity_poly.rcsb_non_std_monomer_count and rcsb_non_std_monomers
# 15-May-2019 jdw add _rcsb_entry_info.na_polymer_entity_types update enumerations for _rcsb_entry_info.selected_polymer_entity_types
# 19-May-2019 jdw add method __getStructConfInfo()
# 21-May-2019 jdw handle odd ordering of records in struct_ref_seq_dif.
# 25-Nov-2019 jdw add method normalizeCitiationJournalAbbrev() and dependencies
#
##
"""
Helper class implements entry-level method references in the RCSB dictionary extension.

All data accessors and structures here refer to dictionary category and attribute names.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

# pylint: disable=too-many-lines

import logging
from string import capwords

from mmcif.api.DataCategory import DataCategory

logger = logging.getLogger(__name__)


def cmpElements(lhs, rhs):
    return 0 if (lhs[-1].isdigit() or lhs[-1] in ["R", "S"]) and rhs[0].isdigit() else -1


class DictMethodEntryHelper(object):
    """Helper class implements entry-level method references in the RCSB dictionary extension."""

    def __init__(self, **kwargs):
        """
        Args:
            resourceProvider: (obj) instance of DictMethodResourceProvider()

        """
        #
        logger.debug("Dictionary entry method helper init with kwargs %r", kwargs)
        self._raiseExceptions = kwargs.get("raiseExceptions", False)
        #
        rP = kwargs.get("resourceProvider")
        self.__commonU = rP.getResource("DictMethodCommonUtils instance") if rP else None
        self.__dApi = rP.getResource("Dictionary API instance (pdbx_core)") if rP else None
        #
        self.__crP = rP.getResource("CitationReferenceProvider instance") if rP else None
        self.__jtaP = rP.getResource("JournalTitleAbbreviationProvider instance") if rP else None
        #
        # logger.debug("Dictionary entry method helper init")

    def echo(self, msg):
        logger.info(msg)

    def deferredItemMethod(self, dataContainer, catName, atName, **kwargs):
        """Placeholder for an item method."""
        _ = kwargs
        logger.debug("Called deferred item method %r %r for %r", catName, atName, dataContainer.getName())
        return True

    def deferredCategoryMethod(self, dataContainer, catName, **kwargs):
        """Placeholder for a category method."""
        _ = kwargs
        logger.debug("Called deferred category method %r for %r", catName, dataContainer.getName())
        return True

    def setDatablockId(self, dataContainer, catName, atName, **kwargs):
        """Item-level method to set the value of the input item to the current container name.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name
            atName (str): Attribute name

        Returns:
            bool: True for success or False otherwise
        """
        logger.debug("Starting catName %s atName %s kwargs %r", catName, atName, kwargs)
        try:
            val = dataContainer.getName()
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=[atName]))
            #
            cObj = dataContainer.getObj(catName)
            if not cObj.hasAttribute(atName):
                cObj.appendAttribute(atName)
            #
            rc = cObj.getRowCount()
            numRows = rc if rc else 1
            for ii in range(numRows):
                cObj.setValue(val, atName, ii)
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def setLoadDateTime(self, dataContainer, catName, atName, **kwargs):
        """Set the value of the input data item with container load date.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name
            atName (str): Attribute name

        Returns:
            bool: True for success or False otherwise
        """
        logger.debug("Starting catName %s atName %s kwargs %r", catName, atName, kwargs)
        try:
            val = dataContainer.getProp("load_date")
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=[atName]))
            #
            cObj = dataContainer.getObj(catName)
            if not cObj.hasAttribute(atName):
                cObj.appendAttribute(atName)
            #
            rc = cObj.getRowCount()
            numRows = rc if rc else 1
            for ii in range(numRows):
                cObj.setValue(val, atName, ii)
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def setLocator(self, dataContainer, catName, atName, **kwargs):
        """Set the value of the input data item with container locator path.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name
            atName (str): Attribute name

        Returns:
            bool: True for success or False otherwise
        """
        logger.debug("Starting catName %s atName %s kwargs %r", catName, atName, kwargs)
        try:
            val = dataContainer.getProp("locator")
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=[atName]))
            #
            cObj = dataContainer.getObj(catName)
            if not cObj.hasAttribute(atName):
                cObj.appendAttribute(atName)
            #
            rc = cObj.getRowCount()
            numRows = rc if rc else 1
            for ii in range(numRows):
                cObj.setValue(val, atName, ii)
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def setRowIndex(self, dataContainer, catName, atName, **kwargs):
        """Set the values of the input data item with the category row index.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name
            atName (str): Attribute name

        Returns:
            bool: True for success or False otherwise
        """
        logger.debug("Starting catName %s atName %s kwargs %r", catName, atName, kwargs)
        try:
            if not dataContainer.exists(catName):
                # exit if there is no category to index
                return False
            #
            cObj = dataContainer.getObj(catName)
            if not cObj.hasAttribute(atName):
                cObj.appendAttribute(atName)
            #
            rc = cObj.getRowCount()
            numRows = rc if rc else 1
            for ii, iRow in enumerate(range(numRows), 1):
                # Note - we set the integer value as a string  -
                cObj.setValue(str(ii), atName, iRow)
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def aggregateCitationAuthors(self, dataContainer, catName, atName, **kwargs):
        """Set the value of the input data item with list of citation authors.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name
            atName (str): Attribute name

        Returns:
            bool: True for success or False otherwise
        """
        logger.debug("Starting catName %s atName %s kwargs %r", catName, atName, kwargs)
        try:
            if not dataContainer.exists(catName) or not dataContainer.exists("citation_author"):
                return False
            #
            cObj = dataContainer.getObj(catName)
            if not cObj.hasAttribute(atName):
                cObj.appendAttribute(atName)
            citIdL = cObj.getAttributeValueList("id")
            #
            tObj = dataContainer.getObj("citation_author")
            #
            citIdL = list(set(citIdL))
            tD = {}
            for ii, citId in enumerate(citIdL):
                tD[citId] = tObj.selectValuesWhere("name", citId, "citation_id")
            for ii in range(cObj.getRowCount()):
                citId = cObj.getValue("id", ii)
                cObj.setValue("|".join(tD[citId]), atName, ii)
            return True
        except Exception as e:
            logger.exception("Failing for %r with %s", dataContainer.getName(), str(e))
        return False

    def normalizeCitationJournalAbbrev(self, dataContainer, catName, atName, **kwargs):
        """Normalize citation journal abbrev.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name
            atName (str): Attribute name

        Returns:
            bool: True for success or False otherwise
        """
        logger.debug("Starting catName %s atName %s kwargs %r", catName, atName, kwargs)
        try:
            if not dataContainer.exists(catName):
                return False
            #
            cObj = dataContainer.getObj(catName)
            if not cObj.hasAttribute("journal_abbrev") or not cObj.hasAttribute("id") or not cObj.hasAttribute("journal_id_ISSN"):
                return False
            #
            if not cObj.hasAttribute(atName):
                cObj.appendAttribute(atName)
            #
            rcsbId = dataContainer.getName()
            for ii in range(cObj.getRowCount()):
                # citId = cObj.getValue("id", ii)
                issn = cObj.getValueOrDefault("journal_id_ISSN", ii, defaultValue=None)
                curAbbrev = cObj.getValueOrDefault("journal_abbrev", ii, defaultValue=None)
                if curAbbrev:
                    revAbbrev = self.__updateJournalAbbreviation(rcsbId, issn, curAbbrev)
                revAbbrev = revAbbrev if revAbbrev else curAbbrev
                #
                logger.debug("%s journal abbreviation issn %r current %r normalized %r", rcsbId, issn, curAbbrev, revAbbrev)
                cObj.setValue(revAbbrev, atName, ii)
            return True
        except Exception as e:
            logger.exception("Failing for %r with %s", dataContainer.getName(), str(e))
        return False

    def __updateJournalAbbreviation(self, rcsbId, issn, curAbbrev):
        try:
            revAbbrev = None
            if issn:
                medlineAbbrev = self.__crP.getMedlineJournalAbbreviation(issn)
                # medlineIsoAbbrev = self.__crP.getMedlineJournalIsoAbbreviation(issn)
                crIssn = issn.replace("-", "")
                crTitle = self.__crP.getCrossRefJournalTitle(crIssn)
                #
                revAbbrev = medlineAbbrev
                if not medlineAbbrev and not crTitle:
                    logger.debug("%s: missing information for issn %r curAbbrev %r", rcsbId, issn, curAbbrev)
                    revAbbrev = capwords(curAbbrev.replace(".", " "))
                elif not medlineAbbrev:
                    revAbbrev = self.__jtaP.getJournalAbbreviation(crTitle, usePunctuation=False)
            else:
                if curAbbrev.upper() in ["TO BE PUBLISHED", "IN PREPARATION"]:
                    revAbbrev = "To be published"
                elif curAbbrev.upper().startswith("THESIS"):
                    revAbbrev = "Thesis"
                else:
                    revAbbrev = capwords(curAbbrev.replace(".", " "))
                    logger.debug("%r: missing issn and non-standard abbrev for %r", rcsbId, curAbbrev)

                if not curAbbrev:
                    logger.info("%r: missing issn and journal abbrev", rcsbId)
                #
            logger.debug("%s: revised: %r current: %r", rcsbId, revAbbrev, curAbbrev)
        except Exception as e:
            logger.exception("Failing on %r %r %r with %r", rcsbId, issn, curAbbrev, str(e))

        return revAbbrev

    def assignPrimaryCitation(self, dataContainer, catName, atName, **kwargs):
        """Normalize citation journal abbrev.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name
            atName (str): Attribute name

        Returns:
            bool: True for success or False otherwise
        """
        logger.debug("Starting catName %s atName %s kwargs %r", catName, atName, kwargs)
        try:
            if not dataContainer.exists(catName):
                return False
            #
            cObj = dataContainer.getObj(catName)
            if not cObj.hasAttribute(atName):
                cObj.appendAttribute(atName)
            #
            for ii in range(cObj.getRowCount()):
                citId = cObj.getValue("id", ii)
                if citId.upper() == "PRIMARY":
                    cObj.setValue("Y", atName, ii)
                else:
                    cObj.setValue("N", atName, ii)
            return True
        except Exception as e:
            logger.exception("Failing for %r with %s", dataContainer.getName(), str(e))
        return False

    def __getEmdbIdentifiers(self, dataContainer):
        """[summary]

        Args:
            dataContainer ([type]): [description]

        Returns:
            [type]: [description]

            #
            loop_
            _database_2.database_id
            _database_2.database_code
            PDB   6QUY
            WWPDB D_1292100913
            EMDB  EMD-4644
            #
            loop_
            _pdbx_database_related.db_name
            _pdbx_database_related.details
            _pdbx_database_related.db_id
            _pdbx_database_related.content_type
            EMDB 'HsCKK (human CAMSAP1) decorated 13pf taxol-GDP microtubule (asymmetric unit)' EMD-4643 'other EM volume'
            PDB  'HsCKK (human CAMSAP1) decorated 13pf taxol-GDP microtubule (asymmetric unit)' 6QUS     unspecified
            EMDB 'NgCKK (N.Gruberi CKK) decorated 13pf taxol-GDP microtubule'                   EMD-4644 'associated EM volume'
            #
        """
        emdbIdD = {}
        emdbIdAltD = {}
        if dataContainer.exists("database_2"):
            dbObj = dataContainer.getObj("database_2")
            for ii in range(dbObj.getRowCount()):
                dbId = dbObj.getValue("database_id", ii)
                dbCode = dbObj.getValue("database_code", ii)
                if dbId.upper() == "EMDB":
                    emdbIdD[dbCode] = "associated EM volume"

        if dataContainer.exists("pdbx_database_related"):
            drObj = dataContainer.getObj("pdbx_database_related")
            for ii in range(drObj.getRowCount()):
                dbCode = drObj.getValue("db_id", ii)
                dbName = drObj.getValue("db_name", ii)
                contentType = drObj.getValue("content_type", ii)
                if dbName.upper() == "EMDB" and contentType.upper() == "ASSOCIATED EM VOLUME" and dbCode not in emdbIdD:
                    emdbIdD[dbCode] = "associated EM volume"
                elif dbName.upper() == "EMDB" and contentType.upper() != "ASSOCIATED EM VOLUME" and dbCode not in emdbIdAltD:
                    emdbIdAltD[dbCode] = contentType
        return emdbIdD, emdbIdAltD

    def buildContainerEntryIds(self, dataContainer, catName, **kwargs):
        """Load the input category with rcsb_entry_container_identifiers content.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise

        For example:

        loop_
        _rcsb_entry_container_identifiers.entry_id
        _rcsb_entry_container_identifiers.entity_ids
        _rcsb_entry_container_identifiers.polymer_entity_ids_polymer
        _rcsb_entry_container_identifiers.non-polymer_entity_ids
        _rcsb_entry_container_identifiers.assembly_ids
        _rcsb_entry_container_identifiers.rcsb_id
        ...

        """
        logger.debug("Starting catName  %s kwargs %r", catName, kwargs)
        try:
            if not dataContainer.exists("entry"):
                return False
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=self.__dApi.getAttributeNameList(catName)))
            #
            cObj = dataContainer.getObj(catName)

            tObj = dataContainer.getObj("entry")
            entryId = tObj.getValue("id", 0)
            cObj.setValue(entryId, "entry_id", 0)
            cObj.setValue(entryId, "rcsb_id", 0)

            #
            tObj = dataContainer.getObj("entity")
            entityIdL = tObj.getAttributeValueList("id")
            cObj.setValue(",".join(entityIdL), "entity_ids", 0)
            #
            #
            tIdL = tObj.selectValuesWhere("id", "polymer", "type")
            tV = ",".join(tIdL) if tIdL else "?"
            cObj.setValue(tV, "polymer_entity_ids", 0)

            tIdL = tObj.selectValuesWhere("id", "non-polymer", "type")
            tV = ",".join(tIdL) if tIdL else "?"
            cObj.setValue(tV, "non-polymer_entity_ids", 0)
            #
            tIdL = tObj.selectValuesWhere("id", "branched", "type")
            tV = ",".join(tIdL) if tIdL else "?"
            cObj.setValue(tV, "branched_entity_ids", 0)
            #
            # tIdL = tObj.selectValuesWhere("id", "water", "type")
            # tV = ",".join(tIdL) if tIdL else "?"
            # cObj.setValue(tV, "water_entity_ids", 0)
            #
            tObj = dataContainer.getObj("pdbx_struct_assembly")
            assemblyIdL = tObj.getAttributeValueList("id") if tObj else []
            tV = ",".join(assemblyIdL) if assemblyIdL else "?"
            cObj.setValue(tV, "assembly_ids", 0)
            #
            #
            emdbIdD, emdbIdAltD = self.__getEmdbIdentifiers(dataContainer)
            tV = ",".join([tId for tId in emdbIdD]) if emdbIdD else "?"
            cObj.setValue(tV, "emdb_ids", 0)
            tV = ",".join([tId for tId in emdbIdAltD]) if emdbIdAltD else "?"
            cObj.setValue(tV, "related_emdb_ids", 0)
            #
            modelIdList = self.__commonU.getModelIdList(dataContainer)
            tV = ",".join([str(tId) for tId in modelIdList]) if modelIdList else "?"
            cObj.setValue(tV, "model_ids", 0)
            #
            return True
        except Exception as e:
            logger.exception("For %s failing with %s", catName, str(e))
        return False

    def consolidateAccessionDetails(self, dataContainer, catName, **kwargs):
        """Consolidate accession details into the rcsb_accession_info category. Also include
        a flag for the availability of any supporting experimental data.

        Args:
            dataContainer (object): mmif.api.DataContainer object instance
            catName (str): Category name

        Returns:
            bool: True for success or False otherwise

        For example:
            For example -
             _rcsb_accession_info.entry_id                1ABC
             _rcsb_accession_info.status_code             REL
             _rcsb_accession_info.deposit_date            2018-01-11
             _rcsb_accession_info.initial_release_date    2018-03-23
             _rcsb_accession_info.major_revision          1
             _rcsb_accession_info.minor_revision          2
             _rcsb_accession_info.revision_date           2018-10-25


            Taking data values from:

            _pdbx_database_status.entry_id                        3OQP
            _pdbx_database_status.deposit_site                    RCSB
            _pdbx_database_status.process_site                    RCSB
            _pdbx_database_status.recvd_initial_deposition_date   2010-09-03
            _pdbx_database_status.status_code                     REL
            _pdbx_database_status.status_code_sf                  REL
            _pdbx_database_status.status_code_mr                  ?
            _pdbx_database_status.status_code_cs                  ?
            _pdbx_database_status.pdb_format_compatible           Y
            _pdbx_database_status.methods_development_category    ?
            _pdbx_database_status.SG_entry                        Y
            #
            loop_
            _pdbx_audit_revision_history.ordinal
            _pdbx_audit_revision_history.data_content_type
            _pdbx_audit_revision_history.major_revision
            _pdbx_audit_revision_history.minor_revision
            _pdbx_audit_revision_history.revision_date
            1 'Structure model' 1 0 2010-10-13
            2 'Structure model' 1 1 2011-07-13
            3 'Structure model' 1 2 2011-07-20
            4 'Structure model' 1 3 2014-11-12
            5 'Structure model' 1 4 2017-10-25
            #

            #  - For EM and SAS -
            _pdbx_database_related.db_name        EMDB
            _pdbx_database_related.details
            'pseudo-atomic model of the RNA polymerase lambda-based antitermination complex solved by cryo-EM'
            _pdbx_database_related.db_id          EMD-3561
            _pdbx_database_related.content_type   'associated EM volume'
        """
        ##
        try:
            logger.debug("Starting with  %r %r %r", dataContainer.getName(), catName, kwargs)
            #
            # if there is incomplete accessioninformation then exit
            if not (dataContainer.exists("pdbx_database_status") or dataContainer.exists("pdbx_audit_revision_history")):
                return False
            # Create the new target category
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=self.__dApi.getAttributeNameList(catName)))

            cObj = dataContainer.getObj(catName)
            #
            tObj = dataContainer.getObj("pdbx_database_status")
            entryId = tObj.getValue("entry_id", 0)
            statusCode = tObj.getValue("status_code", 0)
            depositDate = tObj.getValue("recvd_initial_deposition_date", 0)
            #
            cObj.setValue(entryId, "entry_id", 0)
            cObj.setValue(statusCode, "status_code", 0)
            cObj.setValue(depositDate, "deposit_date", 0)
            # cObj.setValue(depositDate[:4], "deposit_year", 0)
            #
            # -- Experimental data availability --
            #
            expDataRelFlag = "N"
            statusSf = tObj.getValueOrDefault("status_code_sf", 0, defaultValue=None)
            statusMr = tObj.getValueOrDefault("status_code_mr", 0, defaultValue=None)
            statusCs = tObj.getValueOrDefault("status_code_cs", 0, defaultValue=None)
            #
            if statusSf == "REL" or statusMr == "REL" or statusCs == "REL":
                expDataRelFlag = "Y"
            else:
                if dataContainer.exists("pdbx_database_related"):
                    rObj = dataContainer.getObj("pdbx_database_related")
                    ctL = rObj.getAttributeValueList("content_type")
                    if "associated EM volume" in ctL or "associated SAS data" in ctL:
                        expDataRelFlag = "Y"
            #
            cObj.setValue(expDataRelFlag, "has_released_experimental_data", 0)
            #
            tObj = dataContainer.getObj("pdbx_audit_revision_history")
            nRows = tObj.getRowCount()
            # Assuming the default sorting order from the release module -
            releaseDate = tObj.getValue("revision_date", 0)
            minorRevision = tObj.getValue("minor_revision", nRows - 1)
            majorRevision = tObj.getValue("major_revision", nRows - 1)
            revisionDate = tObj.getValue("revision_date", nRows - 1)
            cObj.setValue(releaseDate, "initial_release_date", 0)
            # cObj.setValue(releaseDate[:4], "initial_release_year", 0)
            cObj.setValue(minorRevision, "minor_revision", 0)
            cObj.setValue(majorRevision, "major_revision", 0)
            cObj.setValue(revisionDate, "revision_date", 0)
            #
            return True
        except Exception as e:
            logger.exception("In %s for %s failing with %s", dataContainer.getName(), catName, str(e))
        return False

    def addEntryInfo(self, dataContainer, catName, **kwargs):
        """
        Add  _rcsb_entry_info, for example:
             _rcsb_entry_info.entry_id                      1ABC
             _rcsb_entry_info.polymer_composition           'heteromeric protein'
             _rcsb_entry_info.experimental_method           'multiple methods'
             _rcsb_entry_info.experimental_method_count     2
             _rcsb_entry_info.polymer_entity_count          2
             _rcsb_entry_info.entity_count                  2
             _rcsb_entry_info.nonpolymer_entity_count       2
             _rcsb_entry_info.branched_entity_count         0
             _rcsb_entry_info.software_programs_combined    'Phenix;RefMac'
             ....

        Also add the related field:

        _entity_poly.rcsb_entity_polymer_type

            'Protein'   'polypeptide(D) or polypeptide(L)'
            'DNA'       'polydeoxyribonucleotide'
            'RNA'       'polyribonucleotide'
            'NA-hybrid' 'polydeoxyribonucleotide/polyribonucleotide hybrid'
            'Other'      'polysaccharide(D), polysaccharide(L), cyclic-pseudo-peptide, peptide nucleic acid, or other'
            #
          _rcsb_entry_info.deposited_polymer_monomer_count
          'polymer_entity_count_protein',
          'polymer_entity_count_nucleic_acid',
          'polymer_entity_count_nucleic_acid_hybrid',
          'polymer_entity_count_DNA',
          'polymer_entity_count_RNA',

        """
        try:
            logger.debug("Starting with %r %r %r", dataContainer.getName(), catName, kwargs)
            # Exit if source categories are missing
            if not (dataContainer.exists("exptl") and dataContainer.exists("entity")):
                return False
            #
            # Create the new target category rcsb_entry_info
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=self.__dApi.getAttributeNameList(catName)))
            # --------------------------------------------------------------------------------------------------------
            # catName = rcsb_entry_info
            cObj = dataContainer.getObj(catName)
            #
            # --------------------------------------------------------------------------------------------------------
            #  Filter experimental methods
            #
            xObj = dataContainer.getObj("exptl")
            entryId = xObj.getValue("entry_id", 0)
            methodL = xObj.getAttributeValueList("method")
            methodCount, expMethod = self.__commonU.filterExperimentalMethod(methodL)
            cObj.setValue(entryId, "entry_id", 0)
            cObj.setValue(expMethod, "experimental_method", 0)
            cObj.setValue(methodCount, "experimental_method_count", 0)
            #
            # --------------------------------------------------------------------------------------------------------
            #  Experimental resolution -
            #
            resL = self.__filterExperimentalResolution(dataContainer)
            if resL:
                cObj.setValue(",".join(resL), "resolution_combined", 0)
            #
            # ---------------------------------------------------------------------------------------------------------
            # Consolidate software details -
            #
            swNameL = []
            if dataContainer.exists("software"):
                swObj = dataContainer.getObj("software")
                swNameL.extend(swObj.getAttributeUniqueValueList("name"))
            if dataContainer.exists("pdbx_nmr_software"):
                swObj = dataContainer.getObj("pdbx_nmr_software")
                swNameL.extend(swObj.getAttributeUniqueValueList("name"))
            if dataContainer.exists("em_software"):
                swObj = dataContainer.getObj("em_software")
                swNameL.extend(swObj.getAttributeUniqueValueList("name"))
            if swNameL:
                swNameD = {swName.upper().strip(): True for swName in swNameL if swName not in [".", "?"]}
                swNameL = sorted(swNameD.keys())
                cObj.setValue(";".join(swNameL), "software_programs_combined", 0)
            # ---------------------------------------------------------------------------------------------------------
            #  ENTITY FEATURES
            #
            #  entity and polymer entity counts -
            ##
            eObj = dataContainer.getObj("entity")
            eTypeL = eObj.getAttributeValueList("type")
            #
            numPolymers = 0
            numNonPolymers = 0
            numBranched = 0
            numSolvent = 0
            for eType in eTypeL:
                if eType == "polymer":
                    numPolymers += 1
                elif eType == "non-polymer":
                    numNonPolymers += 1
                elif eType == "branched":
                    numBranched += 1
                elif eType == "water":
                    numSolvent += 1
                else:
                    logger.error("Unexpected entity type for %s %s", dataContainer.getName(), eType)
            totalEntities = numPolymers + numNonPolymers + numBranched + numSolvent
            #
            # Simplified entity polymer type: 'Protein', 'DNA', 'RNA', 'NA-hybrid', or 'Other'
            pTypeL = []
            if dataContainer.exists("entity_poly"):
                epObj = dataContainer.getObj("entity_poly")
                pTypeL = epObj.getAttributeValueList("type")
                #
                atName = "rcsb_entity_polymer_type"
                if not epObj.hasAttribute(atName):
                    epObj.appendAttribute(atName)
                for ii in range(epObj.getRowCount()):
                    epObj.setValue(self.__commonU.filterEntityPolyType(pTypeL[ii]), atName, ii)
            #
            # Add any branched entity types to the type list -
            if dataContainer.exists("pdbx_entity_branch"):
                ebObj = dataContainer.getObj("pdbx_entity_branch")
                pTypeL.extend(ebObj.getAttributeValueList("type"))
            #
            polymerCompClass, ptClass, naClass, eptD = self.__commonU.getPolymerComposition(pTypeL)
            if eptD and len(eptD) > 2:
                logger.info("%s entity type count=%d class=%s typeD %r", dataContainer.getName(), len(eptD), polymerCompClass, eptD)
            #
            cObj.setValue(polymerCompClass, "polymer_composition", 0)
            cObj.setValue(ptClass, "selected_polymer_entity_types", 0)
            cObj.setValue(naClass, "na_polymer_entity_types", 0)
            cObj.setValue(numPolymers, "polymer_entity_count", 0)
            cObj.setValue(numNonPolymers, "nonpolymer_entity_count", 0)
            cObj.setValue(numBranched, "branched_entity_count", 0)
            cObj.setValue(numSolvent, "solvent_entity_count", 0)
            cObj.setValue(totalEntities, "entity_count", 0)
            #
            num = eptD["protein"] if "protein" in eptD else 0
            cObj.setValue(num, "polymer_entity_count_protein", 0)
            #
            num = eptD["NA-hybrid"] if "NA-hybrid" in eptD else 0
            cObj.setValue(num, "polymer_entity_count_nucleic_acid_hybrid", 0)
            #
            numDNA = eptD["DNA"] if "DNA" in eptD else 0
            cObj.setValue(numDNA, "polymer_entity_count_DNA", 0)
            #
            numRNA = eptD["RNA"] if "RNA" in eptD else 0
            cObj.setValue(numRNA, "polymer_entity_count_RNA", 0)
            cObj.setValue(numDNA + numRNA, "polymer_entity_count_nucleic_acid", 0)
            #
            # ---------------------------------------------------------------------------------------------------------
            # INSTANCE FEATURES
            #
            ##
            repModelL = ["1"]
            if self.__commonU.hasMethodNMR(methodL):
                repModelL = self.__getRepresentativeModels(dataContainer)
            logger.debug("Representative model list %r", repModelL)
            #
            instanceTypeCountD = self.__commonU.getInstanceTypeCounts(dataContainer)
            cObj.setValue(instanceTypeCountD["polymer"], "deposited_polymer_entity_instance_count", 0)
            cObj.setValue(instanceTypeCountD["non-polymer"], "deposited_nonpolymer_entity_instance_count", 0)

            #
            # Various atom counts -
            #
            repModelId = repModelL[0]
            numHeavyAtomsModel, numHydrogenAtomsModel, numAtomsTotal, numModelsTotal = self.__commonU.getDepositedAtomCounts(dataContainer, modelId=repModelId)
            #
            logger.debug("numAtomsTotal %d numHeavyAtomsModel %d numModelsTotal %d", numAtomsTotal, numHeavyAtomsModel, numModelsTotal)
            logger.debug("entity type atom counts %r", self.__commonU.getEntityTypeHeavyAtomCounts(dataContainer, modelId=repModelId))
            logger.debug("instance atom counts %r", self.__commonU.getEntityTypeHeavyAtomCounts(dataContainer, modelId=repModelId))
            #

            if numHeavyAtomsModel > 0:
                cObj.setValue(numHeavyAtomsModel, "deposited_atom_count", 0)
                cObj.setValue(numModelsTotal, "deposited_model_count", 0)
                cObj.setValue(numHydrogenAtomsModel, "deposited_hydrogen_atom_count", 0)
                tCD = self.__commonU.getEntityTypeHeavyAtomCounts(dataContainer, modelId=repModelId)
                wCount = tCD["water"] if tCD and "water" in tCD else 0
                cObj.setValue(wCount, "deposited_solvent_atom_count", 0)
            #
            # ---------------------------------------------------------------------------------------------------------
            #  Deposited monomer/residue instance counts
            #
            #  Get modeled and unmodeled residue counts
            #
            modeledCount, unModeledCount = self.__commonU.getDepositedMonomerCounts(dataContainer, modelId=repModelId)
            cObj.setValue(modeledCount, "deposited_modeled_polymer_monomer_count", 0)
            cObj.setValue(unModeledCount, "deposited_unmodeled_polymer_monomer_count", 0)
            cObj.setValue(modeledCount + unModeledCount, "deposited_polymer_monomer_count", 0)
            #
            # ---------------------------------------------------------------------------------------------------------
            #  Counts of intermolecular bonds/linkages
            #
            #
            bCountsD = self.__commonU.getInstanceConnectionCounts(dataContainer)
            cObj.setValue(bCountsD["disulf"], "disulfide_bond_count", 0)
            cObj.setValue(bCountsD["metalc"], "inter_mol_metalic_bond_count", 0)
            cObj.setValue(bCountsD["covale"], "inter_mol_covalent_bond_count", 0)
            #
            cisPeptideD = self.__commonU.getCisPeptides(dataContainer)
            cObj.setValue(len(cisPeptideD), "cis_peptide_count", 0)
            #
            # This is reset in anothor method - filterSourceOrganismDetails()
            cObj.setValue(None, "polymer_entity_taxonomy_count", 0)
            #
            fw = self.__commonU.getFormulaWeightNonSolvent(dataContainer)
            cObj.setValue(str(round(fw, 2)), "molecular_weight", 0)
            #
            # nonpolymer_bound_components
            #
            bcL = self.__commonU.getBoundNonpolymersComponentIds(dataContainer)
            if bcL:
                cObj.setValue(";".join(bcL), "nonpolymer_bound_components", 0)
            #
            # polymer_molecular_weight_minimum
            # polymer_molecular_weight_maximum
            # nonpolymer_molecular_weight_minimum
            # nonpolymer_molecular_weight_maximum
            # branched_molecular_weight_minimum
            # branched_molecular_weight_maximum
            #
            fwBoundD = self.__commonU.getEntityFormulaWeightBounds(dataContainer)
            if "polymer" in fwBoundD and fwBoundD["polymer"]["min"] and fwBoundD["polymer"]["max"]:
                cObj.setValue(str(round(fwBoundD["polymer"]["min"], 2)), "polymer_molecular_weight_minimum", 0)
                cObj.setValue(str(round(fwBoundD["polymer"]["max"], 2)), "polymer_molecular_weight_maximum", 0)
            if "non-polymer" in fwBoundD and fwBoundD["non-polymer"]["min"] and fwBoundD["non-polymer"]["max"]:
                cObj.setValue(str(round(fwBoundD["non-polymer"]["min"], 2)), "nonpolymer_molecular_weight_minimum", 0)
                cObj.setValue(str(round(fwBoundD["non-polymer"]["max"], 2)), "nonpolymer_molecular_weight_maximum", 0)
            if "branched" in fwBoundD and fwBoundD["branched"]["min"] and fwBoundD["branched"]["max"]:
                cObj.setValue(str(round(fwBoundD["branched"]["min"], 2)), "branched_molecular_weight_minimum", 0)
                cObj.setValue(str(round(fwBoundD["branched"]["max"], 2)), "branched_molecular_weight_maximum", 0)
            #
            # polymer_monomer_count_maximum
            # polymer_monomer_count_minimum
            #
            polymerLengthBounds = self.__commonU.getEntityPolymerLengthBounds(dataContainer)
            if polymerLengthBounds:
                cObj.setValue(str(polymerLengthBounds[0]), "polymer_monomer_count_minimum", 0)
                cObj.setValue(str(polymerLengthBounds[1]), "polymer_monomer_count_maximum", 0)
            #
            # ---------------------------------------------------------------------------------------------------------
            # Consolidate diffraction wavelength details -
            wL = []
            try:
                if dataContainer.exists("diffrn_radiation_wavelength"):
                    swObj = dataContainer.getObj("diffrn_radiation_wavelength")
                    wL.extend(swObj.getAttributeUniqueValueList("wavelength"))
                if dataContainer.exists("diffrn_radiation"):
                    swObj = dataContainer.getObj("diffrn_radiation")
                    if swObj.hasAttribute("pdbx_wavelength"):
                        wL.extend(swObj.getAttributeUniqueValueList("pdbx_wavelength"))
                    if swObj.hasAttribute("pdbx_wavelength_list"):
                        tL = []
                        for tS in swObj.getAttributeUniqueValueList("pdbx_wavelength_list"):
                            tL.extend(tS.split(","))
                        if tL:
                            wL.extend(tL)
                if dataContainer.exists("diffrn_source"):
                    swObj = dataContainer.getObj("diffrn_source")
                    if swObj.hasAttribute("pdbx_wavelength"):
                        wL.extend(swObj.getAttributeUniqueValueList("pdbx_wavelength"))
                    if swObj.hasAttribute("pdbx_wavelength_list"):
                        tL = []
                        for tS in swObj.getAttributeUniqueValueList("pdbx_wavelength_list"):
                            tL.extend(tS.split(","))
                        if tL:
                            wL.extend(tL)
                fL = []
                for wS in wL:
                    try:
                        fL.append(float(wS))
                    except Exception:
                        pass
                if fL:
                    cObj.setValue("%.4f" % min(fL), "diffrn_radiation_wavelength_minimum", 0)
                    cObj.setValue("%.4f" % max(fL), "diffrn_radiation_wavelength_maximum", 0)

            except Exception as e:
                logger.exception("%s failing wavelength processing with %s", entryId, str(e))
            #
            # JDW
            self.__updateReflnsResolution(dataContainer)
            return True
        except Exception as e:
            logger.exception("For %s %r failing with %s", dataContainer.getName(), catName, str(e))
        #
        return False

    def filterBlockByMethod(self, dataContainer, blockName, **kwargs):
        """Filter empty placeholder data categories by experimental method."""
        logger.debug("Starting with %r blockName %r kwargs %r", dataContainer.getName(), blockName, kwargs)
        try:
            if not dataContainer.exists("exptl"):
                return False
            #
            xObj = dataContainer.getObj("exptl")
            methodL = xObj.getAttributeValueList("method")
            objNameL = []
            # Test for a diffraction method in the case of multiple methods
            if len(methodL) > 1:
                isXtal = False
                for method in methodL:
                    if method in ["X-RAY DIFFRACTION", "FIBER DIFFRACTION", "POWDER DIFFRACTION", "ELECTRON CRYSTALLOGRAPHY", "NEUTRON DIFFRACTION", "ELECTRON DIFFRACTION"]:
                        isXtal = True
                        break
                if not isXtal:
                    objNameL = ["cell", "symmetry", "refine", "refine_hist", "software", "diffrn", "diffrn_radiation"]
            else:
                #
                mS = methodL[0].upper()
                if mS in ["X-RAY DIFFRACTION", "FIBER DIFFRACTION", "POWDER DIFFRACTION", "ELECTRON CRYSTALLOGRAPHY", "NEUTRON DIFFRACTION", "ELECTRON DIFFRACTION"]:
                    objNameL = []
                elif mS in ["SOLUTION NMR", "SOLID-STATE NMR"]:
                    objNameL = ["cell", "symmetry", "refine", "refine_hist", "software", "diffrn", "diffrn_radiation"]
                elif mS in ["ELECTRON MICROSCOPY", "CRYO-ELECTRON MICROSCOPY"]:
                    objNameL = ["cell", "symmetry", "refine", "refine_hist", "software", "diffrn", "diffrn_radiation"]
                elif mS in ["SOLUTION SCATTERING", "EPR", "THEORETICAL MODEL", "INFRARED SPECTROSCOPY", "FLUORESCENCE TRANSFER"]:
                    objNameL = ["cell", "symmetry", "refine", "refine_hist", "software", "diffrn", "diffrn_radiation"]
                else:
                    logger.error("%s Unexpected method %r", dataContainer.getName(), mS)
            #
            for objName in objNameL:
                dataContainer.remove(objName)
            return True
        except Exception as e:
            logger.exception("For %s failing with %s", dataContainer.getName(), str(e))
        return False

    def filterEnumerations(self, dataContainer, catName, atName, **kwargs):
        """Standardize the item value to conform to enumeration specifications."""
        logger.debug("Starting with %r %r %r %r", dataContainer.getName(), atName, catName, kwargs)
        subD = {("pdbx_reference_molecule", "class"): [("Anti-tumor", "Antitumor")]}
        try:
            if not dataContainer.exists(catName):
                return False
            #
            cObj = dataContainer.getObj(catName)
            if not cObj.hasAttribute(atName):
                return False
            #
            subL = subD[(catName, atName)] if (catName, atName) in subD else []
            #
            for ii in range(cObj.getRowCount()):
                tV = cObj.getValue(atName, ii)
                if tV and tV not in [".", "?"]:
                    for sub in subL:
                        if sub[0] in tV:
                            tV = tV.replace(sub[0], sub[1])
                            cObj.setValue(tV, atName, ii)
            return True
        except Exception as e:
            logger.exception("%s %s %s failing with %s", dataContainer.getName(), catName, atName, str(e))
        return False

    def __getRepresentativeModels(self, dataContainer):
        """Return the list of representative models

        Example:
            #
            _pdbx_nmr_ensemble.entry_id                                      5TM0
            _pdbx_nmr_ensemble.conformers_calculated_total_number            15
            _pdbx_nmr_ensemble.conformers_submitted_total_number             15
            _pdbx_nmr_ensemble.conformer_selection_criteria                  'all calculated structures submitted'
            _pdbx_nmr_ensemble.representative_conformer                      ?
            _pdbx_nmr_ensemble.average_constraints_per_residue               ?
            _pdbx_nmr_ensemble.average_constraint_violations_per_residue     ?
            _pdbx_nmr_ensemble.maximum_distance_constraint_violation         ?
            _pdbx_nmr_ensemble.average_distance_constraint_violation         ?
            _pdbx_nmr_ensemble.maximum_upper_distance_constraint_violation   ?
            _pdbx_nmr_ensemble.maximum_lower_distance_constraint_violation   ?
            _pdbx_nmr_ensemble.distance_constraint_violation_method          ?
            _pdbx_nmr_ensemble.maximum_torsion_angle_constraint_violation    ?
            _pdbx_nmr_ensemble.average_torsion_angle_constraint_violation    ?
            _pdbx_nmr_ensemble.torsion_angle_constraint_violation_method     ?
            #
            _pdbx_nmr_representative.entry_id             5TM0
            _pdbx_nmr_representative.conformer_id         1
            _pdbx_nmr_representative.selection_criteria   'fewest violations'
        """
        repModelL = []
        if dataContainer.exists("pdbx_nmr_representative"):
            tObj = dataContainer.getObj("pdbx_nmr_representative")
            if tObj.hasAttribute("conformer_id"):
                for ii in range(tObj.getRowCount()):
                    nn = tObj.getValue("conformer_id", ii)
                    if nn is not None and nn.isdigit():
                        repModelL.append(nn)

        if dataContainer.exists("pdbx_nmr_ensemble"):
            tObj = dataContainer.getObj("pdbx_nmr_ensemble")
            if tObj.hasAttribute("representative_conformer"):
                nn = tObj.getValue("representative_conformer", 0)
                if nn is not None and nn and nn.isdigit():
                    repModelL.append(nn)
        #
        repModelL = list(set(repModelL))
        if not repModelL:
            logger.debug("Missing representative model data for %s using 1", dataContainer.getName())
            repModelL = ["1"]

        return repModelL

    def __filterExperimentalResolution(self, dataContainer):
        """Collect resolution estimates from method specific sources."""
        rL = []
        if dataContainer.exists("refine"):
            tObj = dataContainer.getObj("refine")
            if tObj.hasAttribute("ls_d_res_high"):
                for ii in range(tObj.getRowCount()):
                    rv = tObj.getValue("ls_d_res_high", ii)
                    if self.__commonU.isFloat(rv):
                        rL.append(rv)

        if dataContainer.exists("em_3d_reconstruction"):
            tObj = dataContainer.getObj("em_3d_reconstruction")
            if tObj.hasAttribute("resolution"):
                for ii in range(tObj.getRowCount()):
                    rv = tObj.getValue("resolution", ii)
                    if self.__commonU.isFloat(rv):
                        rL.append(rv)
        return rL

    def addCategoryPrimaryCitation(self, dataContainer, blockName, **kwargs):
        """
        Add  rcsb_primary_citation category as a copy or the citation category
        with rcsb extensions.
        """
        try:
            logger.debug("Starting with %r %r %r", dataContainer.getName(), blockName, kwargs)
            # Exit if source categories are missing
            if not dataContainer.exists("citation"):
                return False
            cObj = dataContainer.getObj("citation")
            catName = "rcsb_primary_citation"
            #
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=self.__dApi.getAttributeNameList(catName)))
            # --------------------------------------------------------------------------------------------------------
            rObj = dataContainer.getObj(catName)
            atNameList = self.__dApi.getAttributeNameList(catName)
            logger.debug("Category %s dict attributes %r", catName, atNameList)
            #
            for ii in range(cObj.getRowCount()):
                pv = cObj.getValue("id", ii)
                if pv.upper() == "PRIMARY":
                    for atName in atNameList:
                        if cObj.hasAttribute(atName):
                            rObj.setValue(cObj.getValue(atName, ii), atName, 0)

            return True
        except Exception as e:
            logger.exception("%s %s failing with %s", dataContainer.getName(), catName, str(e))
        return False

    def __updateReflnsResolution(self, dataContainer):
        """Find a plausable data collection diffraction high resolution limit from one of the following sources.
        #
        _rcsb_entry_info.diffrn_resolution_high_value
        _rcsb_entry_info.diffrn_resolution_high_provenance_source

        Update category 'reflns' with any missing resolution extrema data using limits in category reflns_shell.

            _reflns.entry_id                     2DCG
            _reflns.d_resolution_high            0.900
            _reflns.pdbx_diffrn_id               1
            _reflns.pdbx_ordinal                 1


            _refine.entry_id                                 2DCG
            _refine.ls_number_reflns_obs                     15000
            _refine.ls_number_reflns_all                     ?
            _refine.pdbx_ls_sigma_I                          2.000
            _refine.ls_d_res_low                             ?
            _refine.ls_d_res_high                            0.900
            _refine.pdbx_refine_id                           'X-RAY DIFFRACTION'
            _refine.pdbx_diffrn_id                           1

            _reflns_shell.d_res_high             1.18
            _reflns_shell.d_res_low              1.25
            _reflns_shell.pdbx_ordinal           1
            _reflns_shell.pdbx_diffrn_id         1
            #

        """
        try:
            logger.debug("Starting with %r", dataContainer.getName())
            #
            if not dataContainer.exists("exptl") or not dataContainer.exists("rcsb_entry_info"):
                return False
            # --------------------------------------------------------------------------------------------------------
            #  Only applicable to X-ray
            #
            xObj = dataContainer.getObj("exptl")
            methodL = xObj.getAttributeValueList("method")
            _, expMethod = self.__commonU.filterExperimentalMethod(methodL)
            if expMethod not in ["X-ray", "Neutron", "Multiple methods"]:
                return False
            #
            resValue = resProvSource = None
            #
            # Here are the various cases -
            if dataContainer.exists("reflns"):
                rObj = dataContainer.getObj("reflns")
                if rObj.hasAttribute("d_resolution_high"):
                    rvL = rObj.getAttributeValueList("d_resolution_high")
                    fvL = [float(rv) for rv in rvL if self.__commonU.isFloat(rv)]
                    if fvL:
                        resValue = round(min(fvL), 2)
                        resProvSource = "Depositor assigned"

            if not resValue and dataContainer.exists("reflns_shell"):
                rObj = dataContainer.getObj("reflns_shell")
                if rObj.hasAttribute("d_res_high"):
                    rvL = rObj.getAttributeValueList("d_res_high")
                    fvL = [float(rv) for rv in rvL if self.__commonU.isFloat(rv)]
                    if fvL:
                        resValue = round(min(fvL), 2)
                        resProvSource = "From the high resolution shell"

            if not resValue and dataContainer.exists("refine"):

                rObj = dataContainer.getObj("refine")
                if rObj.hasAttribute("ls_d_res_high"):
                    fvL = []
                    for ii in range(rObj.getRowCount()):
                        rId = rObj.getValue("pdbx_refine_id", ii)
                        if rId in ["X-RAY DIFFRACTION", "NEUTRON DIFFRACTION", "FIBER DIFFRACTION"]:
                            rv = rObj.getValue("ls_d_res_high", ii)
                            if self.__commonU.isFloat(rv):
                                fvL.append(float(rv))
                    if fvL:
                        resValue = round(min(fvL), 2)
                        resProvSource = "From refinement resolution cutoff"
            #
            if not resValue:
                logger.debug("No source of data collection resolution available for %r", dataContainer.getName())
            else:
                logger.debug("Data collection diffraction limit %r PS %r", resValue, resProvSource)

            if resValue:
                eObj = dataContainer.getObj("rcsb_entry_info")
                for atName in ["diffrn_resolution_high_value", "diffrn_resolution_high_provenance_source"]:
                    if not eObj.hasAttribute(atName):
                        eObj.appendAttribute(atName)
                eObj.setValue(resValue, "diffrn_resolution_high_value", 0)
                eObj.setValue(resProvSource, "diffrn_resolution_high_provenance_source", 0)
                # --------------------------------------------------------------------------------------------------------
                return True
        except Exception as e:
            logger.exception("%s failing with %s", dataContainer.getName(), str(e))
        return False



##
# File:   DataTypeApiProvider.py
# Author:  J. Westbrook
# Date:   18-Aug-2019
# Version: 0.001 Initial version
#
# Updates:
#
#
##
"""
Data type application and instance information provider.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os

from rcsb.db.define.DataTypeApplicationInfo import DataTypeApplicationInfo
from rcsb.db.define.DataTypeInstanceInfo import DataTypeInstanceInfo
from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.SingletonClass import SingletonClass

logger = logging.getLogger(__name__)


class DataTypeApiProvider(SingletonClass):
    """ Data type application and instance information provider.
    """

    def __init__(self, cfgOb, cachePath, useCache=True, **kwargs):
        """Data type application and instance information provider.

        Args:
            cfgOb (object):  ConfigInfo() object instance
            cachePath (str): path to hold the cache directory
            useCache (bool, optional): flag to use cached files. Defaults to True.

        """
        self.__cfgOb = cfgOb
        self.__configName = self.__cfgOb.getDefaultSectionName()
        self.__useCache = useCache
        self.__cachePath = cachePath
        # self.__contentInfoConfigName = "content_info_helper_configuration"
        self.__fileU = FileUtil()
        self.__contentDefHelper = self.__cfgOb.getHelper("CONTENT_DEF_HELPER_MODULE", sectionName=self.__configName, cfgOb=self.__cfgOb)
        self.__dirPath = os.path.join(cachePath, self.__cfgOb.get("DATA_TYPE_INFO_CACHE_DIR", sectionName=self.__configName))
        self.__kwargs = kwargs
        #
        logger.debug("Leaving constructor")

    def getDataTypeInstanceApi(self, databaseName, **kwargs):
        """Return instance of DataTypeInstanceInfo().

        Args:
            databaseName (str): database name

        Returns:
            (object): Instance of DataTypeInstanceInfo()
        """
        _ = kwargs
        dataTypeInstanceLocatorPath = self.__cfgOb.getPath("INSTANCE_DATA_TYPE_INFO_LOCATOR_PATH", sectionName=self.__configName)
        dataTypeInstanceFile = self.__contentDefHelper.getDataTypeInstanceFile(databaseName) if self.__contentDefHelper else None
        if dataTypeInstanceLocatorPath and dataTypeInstanceFile:
            loc = os.path.join(dataTypeInstanceLocatorPath, dataTypeInstanceFile)
            filePath = self.__reload(loc, self.__dirPath, useCache=self.__useCache)
            dtApi = DataTypeInstanceInfo(filePath)
        else:
            # DataTypeInstanceInfo() provides an internal by-pass mode where no coverage data is available.
            dtApi = DataTypeInstanceInfo(None)
            logger.debug("No data coverage available for database %s", databaseName)
        return dtApi

    def getDataTypeApplicationApi(self, appName, **kwargs):
        """Return instance of DataTypeApplicationInfo.

        Args:
            appName (str): application name (e.g., SQL, ANY)

        Returns:
            (object): Instance of DataTypeApplicationInfo()
        """
        _ = kwargs
        dataTypeApplicationLocator = self.__cfgOb.getPath("APP_DATA_TYPE_INFO_LOCATOR", sectionName=self.__configName)
        filePath = self.__reload(dataTypeApplicationLocator, self.__dirPath, useCache=self.__useCache)
        dtApi = DataTypeApplicationInfo(filePath, dataTyping=appName, workPath=self.__dirPath) if filePath else None
        return dtApi

    def __reload(self, urlTarget, dirPath, useCache=True):
        #
        fn = self.__fileU.getFileName(urlTarget)
        filePath = os.path.join(dirPath, fn)
        logger.debug("Using cache path %s", dirPath)
        self.__fileU.mkdir(dirPath)
        if not useCache:
            try:
                os.remove(filePath)
            except Exception:
                pass
        #
        if useCache and self.__fileU.exists(filePath):
            ok = True
        else:
            logger.debug("Fetch data from source %s", urlTarget)
            ok = self.__fileU.get(urlTarget, os.path.join(dirPath, fn))

        return filePath if ok else None



##
# File:    CaseNormalizedDict.py
# Author:  J. Westbrook
# Date:    4-Sep-2018
# Version: 0.001
#
# Updates:
##
"""
Dictionary container with case insensitive key comparison preserving the
case of the input keys.
"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import collections
import logging

logger = logging.getLogger(__name__)


class CaseNormalizedDict(dict):

    """ Dictionary container with case insensitive key comparison preserving the
        case of the input keys.

    """

    # This example subclasses dict directly following the example from the requests module.
    # proxy = {}

    def __init__(self, data):
        super(CaseNormalizedDict, self).__init__(data)
        self.__local = dict((k.lower(), k) for k in data)
        # for k in data:
        #    self[k] = data[k]

    def __contains__(self, k):
        return k.lower() in self.__local

    def __delitem__(self, k):
        key = self.__local[k.lower()]
        super(CaseNormalizedDict, self).__delitem__(key)
        del self.__local[k.lower()]

    def __getitem__(self, k):
        key = self.__local[k.lower()]
        return super(CaseNormalizedDict, self).__getitem__(key)

    def get(self, k, default=None):
        return self[k] if k in self else default

    def __setitem__(self, k, v):
        super(CaseNormalizedDict, self).__setitem__(k, v)
        self.__local[k.lower()] = k


class CaseNormalizedDict2(collections.MutableMapping):

    """ Dictionary container with case insensitive key comparison preserving the
        case of the input keys.

    """

    # This example subclasses the abstract base class MutableMapping.

    def __init__(self, *args, **kwargs):
        self.__dict__.update(*args, **kwargs)
        self.__local = dict((k.lower(), k) for k in self.__dict__)

    def __contains__(self, k):
        return k.lower() in self.__local

    def __len__(self):
        return len(self.__local)

    def __iter__(self):
        return iter(self.__local)

    def __getitem__(self, k):
        return self.__dict__[self.__local[k.lower()]]

    def __setitem__(self, k, val):
        self.__dict__[k] = val
        self.__local[k.lower()] = k

    def __delitem__(self, k):
        del self.__dict__[k]
        del self.__local[k.lower()]

    def pop(self, k):
        kp = self.__local.pop(k.lower())
        return self.__dict__.pop(kp)

    def getKey(self, k):
        return self.__local.get(k.lower())

    #
    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return "{}, ({})".format(super(CaseNormalizedDict2, self).__repr__(), self.__dict__)



##
# File:    SchemaDocumentHelper.py
# Author:  J. Westbrook
# Date:    7-Jun-2018
# Version: 0.001 Initial version
#
# Updates:
#  22-Jun-2018 jdw change collection attribute id specification to dot notation
#  14-Aug-2018 jdw generalize document key attribute to attribute list
#  20-Aug-2018 jdw slice details added to __schemaContentFilters
#   8-Oct-2018 jdw added getSubCategoryAggregates() method
#   3-Dec-2018 jdw add method getDocumentIndices()
#  16-Jan-2019 jdw add method getDocumentReplaceAttributeNames()
#  11-Mar-2019 jdw add methods getSubCategoryAggregateFeatures() and  getSubCategoryAggregateUnitCardinality()
#  13-Mar-2019 jdw add getCollectionVersion() and getCollectionInfo() and remove getCollections().
#   6-Sep-2019 jdw incorporate search type and brief descriptions
#  23-Oct-2019 jdw add collection subcategory nested property support
#
##
"""
Inject additional document information into a schema definition.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

logger = logging.getLogger(__name__)


class DocumentDefinitionHelper(object):
    """Inject additional configuration information into a document schema definition.

    Single source of document schema semantic configuration content.
    """

    def __init__(self, **kwargs):
        """
        Args:
            cfgOb (obj): instance of ConfigInfo()
            config_section (str): target configuration section name

        """
        # ----
        #
        self.__cfgOb = kwargs.get("cfgOb", None)
        sectionName = kwargs.get("config_section", "document_helper_configuration")
        self.__cfgD = self.__cfgOb.exportConfig(sectionName=sectionName)
        self.__searchTypeD = {}
        self.__searchTypeAttributeD = {}
        self.__attributeDescriptionD = {}
        self.__categoryNested = {}
        self.__subCategoryNested = {}
        self.__attributeSeachPriority = {}
        #
        self.__searchGroupD = {}
        self.__searchGroupAttributeD = {}
        #
        # ----

    def getCollectionInfo(self, schemaName):
        """Returns a list of [{NAME: xx, VERSION: xxx}, ...] for the input schema."""
        cL = []
        try:
            cL = [td for td in self.__cfgD["document_collection_names"][schemaName]]
        except Exception as e:
            logger.debug("Schema definitions name %s failing with %s", schemaName, str(e))
        return cL

    def getCollectionVersion(self, schemaName, collectionName):
        """Return the version string for the the input schema/collection"""
        v = None
        try:
            for td in self.__cfgD["document_collection_names"][schemaName]:
                if collectionName == td["NAME"]:
                    return td["VERSION"]
        except Exception as e:
            logger.debug("Schema definitiona name %s failing with %s", schemaName, str(e))
        return v

    def getExcluded(self, collectionName):
        """For input collection, return the list of excluded schema identifiers."""
        includeL = []
        try:
            includeL = [tS.upper() for tS in self.__cfgD["document_collection_content_filters"][collectionName]["EXCLUDE"]]
        except Exception as e:
            logger.debug("Collection %s failing with %s", collectionName, str(e))
        return includeL

    def getIncluded(self, collectionName):
        """For input collection, return the list of included schema identifiers."""
        excludeL = []
        try:
            excludeL = [tS.upper() for tS in self.__cfgD["document_collection_content_filters"][collectionName]["INCLUDE"]]
        except Exception as e:
            logger.debug("Collection %s failing with %s", collectionName, str(e))
        return excludeL

    def getSliceFilter(self, collectionName):
        """For input collection, return an optional slice filter or None."""
        sf = None
        try:
            sf = self.__cfgD["document_collection_content_filters"][collectionName]["SLICE"]
        except Exception as e:
            logger.debug("Collection %s failing with %s", collectionName, str(e))
        return sf

    def getDocumentExcludedAttributes(self, collectionName, asTuple=True):
        atExcludeD = {}
        for cn, cDL in self.__cfgD["collection_attribute_content_filters"].items():
            if cn != collectionName:
                continue
            for cD in cDL:
                catName = cD["CATEGORY_NAME"]
                if "ATTRIBUTE_NAME_LIST" in cD:
                    for atName in cD["ATTRIBUTE_NAME_LIST"]:
                        if asTuple:
                            atExcludeD[(catName, atName)] = collectionName
                        else:
                            atExcludeD.setdefault(catName, []).append(atName)
        return atExcludeD

    def getDocumentKeyAttributeNames(self, collectionName):
        ret = []
        try:
            for dD in self.__cfgD["collection_indices"][collectionName]:
                if dD["INDEX_NAME"] == "primary":
                    ret = dD["ATTRIBUTE_NAMES"]
                    break
        except Exception as e:
            logger.exception("Collection %s failing with %s", collectionName, str(e))
        return ret

    def getDocumentReplaceAttributeNames(self, collectionName):
        """Return index labeled replace in provided or 'primary' otherwise"""
        ret = []
        try:
            for dD in self.__cfgD["collection_indices"][collectionName]:
                if dD["INDEX_NAME"] == "replace":
                    ret = dD["ATTRIBUTE_NAMES"]
                    break
            if ret:
                return ret
            #
            for dD in self.__cfgD["collection_indices"][collectionName]:
                if dD["INDEX_NAME"] == "primary":
                    ret = dD["ATTRIBUTE_NAMES"]
                    break
        except Exception as e:
            logger.exception("Collection %s failing with %s", collectionName, str(e))
        return ret

    def getDocumentIndices(self, collectionName):
        ret = []
        try:
            ret = [d for d in self.__cfgD["collection_indices"][collectionName] if d["ATTRIBUTE_NAMES"] and len(d["ATTRIBUTE_NAMES"]) > 0]
        except Exception as e:
            logger.exception("Collection %s failing with %s", collectionName, str(e))
        return ret

    def getDocumentIndexAttributes(self, collectionName, indexName):
        ret = []
        try:
            for dD in self.__cfgD["collection_indices"][collectionName]:
                if dD["INDEX_NAME"] == indexName:
                    ret = dD["ATTRIBUTE_NAMES"]
                    break
        except Exception as e:
            logger.exception("Collection %s %s failing with %s", collectionName, indexName, str(e))
        return ret

    def getPrivateDocumentAttributes(self, collectionName):
        ret = []
        try:
            return [d for d in self.__cfgD["collection_private_keys"][collectionName] if d["PRIVATE_DOCUMENT_NAME"] and len(d["PRIVATE_DOCUMENT_NAME"]) > 0]
        except Exception as e:
            logger.debug("Collection %s failing with %s", collectionName, str(e))
        return ret

    def getSubCategoryAggregates(self, collectionName):
        ret = []
        try:
            return [tS["NAME"] for tS in self.__cfgD["collection_subcategory_aggregates"][collectionName]]
        except Exception as e:
            logger.debug("Collection %s failing with %s", collectionName, str(e))
        return ret

    def getSubCategoryAggregateUnitCardinality(self, collectionName, subCategoryName):
        ret = False
        try:
            if collectionName in self.__cfgD["collection_subcategory_aggregates"]:
                for dD in self.__cfgD["collection_subcategory_aggregates"][collectionName]:
                    if dD["NAME"] == subCategoryName:
                        ret = dD["HAS_UNIT_CARDINALITY"]
                        break
        except Exception as e:
            logger.debug("Collection %s failing with %s", collectionName, str(e))
        return ret

    def getSubCategoryAggregateFeatures(self, collectionName):
        ret = []
        try:
            return [tD for tD in self.__cfgD["collection_subcategory_aggregates"][collectionName]]
        except Exception as e:
            logger.debug("Collection %s failing with %s", collectionName, str(e))
        return ret

    def getRetainSingletonObjects(self, collectionName):
        """By default singleton objects are expanded in global scope.  To avoid
        this behaviour set the retain singleton option for the collection.
        """
        ret = False
        try:
            return self.__cfgD["collection_retain_singleton"][collectionName]
        except Exception as e:
            logger.debug("Collection %s failing with %s", collectionName, str(e))
        return ret

    def getSuppressedCategoryRelationships(self, collectionName):
        """
        Example:

           collection_suppress_category_relationships:
               ihm_dev:
                   - PARENT_CATEGORY_NAME: chem_comp
                     CHILD_CATEGORY_NAME: atom_site
                   - PARENT_CATEGORY_NAME: entity_poly_seq
                     CHILD_CATEGORY_NAME: atom_site
        """
        rL = []
        try:
            rL = [tD for tD in self.__cfgD["collection_suppress_category_relationships"][collectionName]]
        except Exception as e:
            logger.debug("Collection %s failing with %s", collectionName, str(e))

        return rL

    def __prepareAttributeSearchContexts(self):
        """
        Example:

        collection_attribute_search_contexts:
            pdbx_core_entity_instance:
                - SEARCH_TYPE: exact-match
                  ATTRIBUTE_NAMES:
                  - rcsb_polymer_instance_feature.name
                - SEARCH_TYPE: default-match
                  ATTRIBUTE_NAMES:
                  - rcsb_entity_instance_domain_scop.domain_class_lineage.name
                  - rcsb_entity_instance_domain_scop.domain_class_lineage.id
                  - rcsb_entity_instance_domain_cath.domain_class_lineage.name
                  - rcsb_entity_instance_domain_cath.domain_class_lineage.id

        returns:
            dict : {collectionName: {(category, attribute): [search type, ...], }, }
            dict : {(category, attribute): [(searchType, collectionName), ... ]

        """
        cD = {}
        pD = {}
        try:
            # preprocess search context data --
            for collectionName, tDL in self.__cfgD["collection_attribute_search_contexts"].items():
                aD = {}
                for tD in tDL:
                    for atName in tD["ATTRIBUTE_NAMES"]:
                        ff = atName.split(".")
                        if len(ff) > 2:
                            logger.error("Bad attribute name for search type %r", atName)
                            continue
                        aD.setdefault((ff[0], ff[1]), []).append(tD["SEARCH_TYPE"])
                        pD.setdefault((ff[0], ff[1]), []).append((tD["SEARCH_TYPE"], collectionName))
                cD[collectionName] = {tup: self.__filterSearchContexts(sorted(list(set(sL)))) for tup, sL in aD.items()}

        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return cD, pD

    def __filterSearchContexts(self, stL, overlapFlag=False):
        """Automatically filter dependent search contexts.

        Leaving this in for now with a disable flab

        Args:
            stL (list): list of search context names
            overlapFlag (bool, optional): disable flag to filter for overlapping search contexts. Defaults to False.

        Returns:
            [type]: [description]
        """
        if overlapFlag:
            if "exact-match" in stL and "full-text" in stL:
                stL.remove("full-text")
        return stL

    def getAttributeSearchContexts(self, collectionName, categoryName, attributeName):
        """Return the list of search types assigned to the input collection/item.

        returns:
            list : [search type, ...]

        """
        rL = []
        try:
            if not self.__searchTypeD:
                self.__searchTypeD, self.__searchTypeAttributeD = self.__prepareAttributeSearchContexts()
            rL = self.__searchTypeD[collectionName][(categoryName, attributeName)]
            # logger.info("Collection %r categoryName %r attributeName %r failing with %s", collectionName, categoryName, attributeName, rL)
        except Exception as e:
            logger.debug(" ---- Collection %sr categoryName %r attributeName %r failing with %s", collectionName, categoryName, attributeName, str(e))

        return rL

    def getSearchContexts(self, categoryName, attributeName):
        """Return the list of search types and collections assigned to the input category/attribute

        Returns:
            list : [(search type, collectionName),  ...]

        """
        try:
            if not self.__searchTypeD:
                self.__searchTypeD, self.__searchTypeAttributeD = self.__prepareAttributeSearchContexts()
            return self.__searchTypeAttributeD.get((categoryName, attributeName), [])
        except Exception as e:
            logger.debug(" ---- CategoryName %r attributeName %r failing with %s", categoryName, attributeName, str(e))
        return []

    def getAllAttributeSearchContexts(self):
        """Return search context data structure

        Returns:
            dict - {(category,attribute): [contexts,,,]}

        """
        try:
            if not self.__searchTypeD:
                self.__searchTypeD, self.__searchTypeAttributeD = self.__prepareAttributeSearchContexts()
            return self.__searchTypeAttributeD
        except Exception as e:
            logger.debug("Failing with %s", str(e))
        return {}

    def isTextSearchType(self, categoryName, attributeName):
        _ = categoryName
        _ = attributeName
        return False

    def __prepareAttributeDescriptions(self):
        """
        Example:

            attribute_descriptions:
                - ATTRIBUTE_NAME: rcsb_entry_container_identifiers.entry_id
                  TYPE: brief
                  TEXT: PDB ID(s)
                - ATTRIBUTE_NAME: pdbx_deposit_group.group_id
                  TYPE: brief
                  TEXT: Deposit Group ID(s)
        """
        aD = {}
        # preprocess description data --
        for tD in self.__cfgD["attribute_descriptions"]:
            atName = tD["ATTRIBUTE_NAME"]
            dType = tD["TYPE"]
            ff = atName.split(".")
            if len(ff) != 2:
                logger.error("Bad attribute name for text description %r", atName)
                continue
            aD[(ff[0], ff[1], dType)] = tD["TEXT"]
        return aD

    def getAttributeDescription(self, categoryName, attributeName, contextType="brief"):
        ret = None
        try:
            self.__attributeDescriptionD = self.__prepareAttributeDescriptions() if not self.__attributeDescriptionD else self.__attributeDescriptionD
            ret = self.__attributeDescriptionD[(categoryName, attributeName, contextType)]
        except Exception as e:
            logger.debug("CategoryName %r attributeName %r failing  %s", categoryName, attributeName, str(e))
        return ret

    #
    def __prepareCategoryNested(self):
        """
        Example:

        collection_category_nested:
            pdbx_core_polymer_entity:
                - CATEGORY: rcsb_polymer_entity_feature_summary
                    NAME: feature_summary
                    CONTEXT_ATTRIBUTE_NAMES:
                    - rcsb_polymer_entity_feature_summary.type
                - CATEGORY: rcsb_polymer_entity_annotation
                    NAME: annotation_type
                    CONTEXT_ATTRIBUTE_NAMES:
                    - rcsb_polymer_entity_annotation.type

                    # new
                    CONTEXT_ATTRIBUTE_VALUES:
                    - CONTEXT_VALUE: a_value
                      SEARCH_PATHS:
                      - a.b
                      - c.d

            # From above ---
            "rcsb_nested_indexing": true,
            "rcsb_nested_indexing_context": [
                {
                "category_name": "annotation_type",
                "category_path": "rcsb_polymer_entity_annotation.type"
                "context_attributes": [
                    {
                        "context_value": 'vxxxxx',
                        "search_paths": [ 'p1', 'p2'],
                        #
                        # new in 2020-10-05
                        "attributes": [
                            {
                                "path": 'p1',
                                "examples": ['ex1', 'ex2']
                            },
                            {
                                "path": 'p2',
                                "examples": ['ex1', 'ex2']
                            }
                        ]
                    }

                ]
                }
            ]

        """
        cD = {}
        try:
            # preprocess the nesting data --
            for collectionName, nDL in self.__cfgD["collection_category_nested"].items():
                catD = {}
                for nD in nDL:
                    cVDL = []
                    if "CONTEXT_ATTRIBUTE_VALUES" in nD:
                        for tD in nD["CONTEXT_ATTRIBUTE_VALUES"]:
                            # To be deprecated
                            if "CONTEXT_VALUE" in tD and "SEARCH_PATHS" in tD and "ATTRIBUTES" not in tD:
                                cVDL.append({"CONTEXT_VALUE": tD["CONTEXT_VALUE"], "SEARCH_PATHS": tD["SEARCH_PATHS"], "ATTRIBUTES": []})
                            # Use this
                            elif "CONTEXT_VALUE" in tD and "ATTRIBUTES" in tD:
                                if "SEARCH_PATHS" in tD:
                                    cVDL.append({"CONTEXT_VALUE": tD["CONTEXT_VALUE"], "SEARCH_PATHS": tD["SEARCH_PATHS"], "ATTRIBUTES": tD["ATTRIBUTES"]})
                                else:
                                    cVDL.append({"CONTEXT_VALUE": tD["CONTEXT_VALUE"], "SEARCH_PATHS": [], "ATTRIBUTES": tD["ATTRIBUTES"]})

                    if "CONTEXT_ATTRIBUTE_NAMES" in nD and "NAME" in nD:
                        catD[nD["CATEGORY"]] = {
                            "CONTEXT_NAME": nD["NAME"],
                            "CONTEXT_PATHS": nD["CONTEXT_ATTRIBUTE_NAMES"],
                            "FIRST_CONTEXT_PATH": nD["CONTEXT_ATTRIBUTE_NAMES"][0],
                            "CONTEXT_ATTRIBUTE_VALUES": cVDL,
                        }
                    elif "NAME" in nD:
                        catD[nD["CATEGORY"]] = {"CONTEXT_NAME": nD["NAME"], "CONTEXT_PATHS": [], "FIRST_CONTEXT_PATH": None, "CONTEXT_ATTRIBUTE_VALUES": cVDL}
                cD[collectionName] = catD
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return cD

    def isCategoryNested(self, collectionName, categoryName):
        """Return is the input category in this collection is nested.

        Args:
            collectionName (str): collection name
            categoryName (str): category name

        Returns:
            (bool): True if nested or False otherwise

        """
        ret = False
        try:
            self.__categoryNested = self.__prepareCategoryNested() if not self.__categoryNested else self.__categoryNested
            ret = categoryName in self.__categoryNested[collectionName]
        except Exception:
            pass
        return ret

    def getCategoryNestedContext(self, collectionName, categoryName):
        """Return is the input category in this collection is nested.

        Args:
            collectionName (str): collection name
            categoryName (str): category name

        Returns:
            dict: {"CONTEXT_NAME": <name>, "CONTEXT_PATHS": <full_path_list>}

        """
        ret = {}
        try:
            self.__categoryNested = self.__prepareCategoryNested() if not self.__categoryNested else self.__categoryNested
            ret = self.__categoryNested[collectionName][categoryName]
        except Exception:
            pass
        return ret

    def getNestedContexts(self, categoryName):
        """Return collections in which the input category is nested.

        Args:
            categoryName (str): category name

        Returns:
            dict: [{"CONTEXT_NAME": <name>, "CONTEXT_PATHS": <full_path_list>}]

        """
        retL = []
        try:
            self.__categoryNested = self.__prepareCategoryNested() if not self.__categoryNested else self.__categoryNested
            for collectionName in self.__categoryNested:
                if categoryName in self.__categoryNested[collectionName]:
                    retL.append(self.__categoryNested[collectionName][categoryName])
        except Exception:
            pass
        return retL

    def __prepareSubCategoryNested(self):
        """
        Example:
        #
        collection_subcategory_nested:
            bird_chem_comp_core:
                - CATEGORY: rcsb_chem_comp_related
                  SUBCATEGORY: resource_lineage
                  CONTEXT_ATTRIBUTE_NAMES:
                  - rcsb_chem_comp_related.resource_lineage_depth
            pdbx_core_polymer_entity:
                - CATEGORY: rcsb_polymer_entity
                  SUBCATEGORY: rcsb_ec_lineage
                - CATEGORY: rcsb_polymer_entity
                  SUBCATEGORY: rcsb_enzyme_class_combined

        """
        cD = {}
        # preprocess the nesting data --
        for collectionName, nDL in self.__cfgD["collection_subcategory_nested"].items():
            subCatD = {}
            for nD in nDL:
                cVDL = []
                if "CONTEXT_ATTRIBUTE_VALUES" in nD:
                    for tD in nD["CONTEXT_ATTRIBUTE_VALUES"]:
                        # to be deprecated
                        if "CONTEXT_VALUE" in tD and "SEARCH_PATHS" in tD and "ATTRIBUTES" not in tD:
                            cVDL.append({"CONTEXT_VALUE": tD["CONTEXT_VALUE"], "SEARCH_PATHS": tD["SEARCH_PATHS"], "ATTRIBUTES": []})
                        # Use this
                        elif "CONTEXT_VALUE" in tD and "ATTRIBUTES" in tD:
                            if "SEARCH_PATHS" in tD:
                                cVDL.append({"CONTEXT_VALUE": tD["CONTEXT_VALUE"], "SEARCH_PATHS": tD["SEARCH_PATHS"], "ATTRIBUTES": tD["ATTRIBUTES"]})
                            else:
                                cVDL.append({"CONTEXT_VALUE": tD["CONTEXT_VALUE"], "SEARCH_PATHS": [], "ATTRIBUTES": tD["ATTRIBUTES"]})
                #
                if "CONTEXT_ATTRIBUTE_NAMES" in nD and "SUBCATEGORY" in nD:
                    subCatD[(nD["CATEGORY"], nD["SUBCATEGORY"])] = {
                        "CONTEXT_NAME": nD["SUBCATEGORY"],
                        "CONTEXT_PATHS": nD["CONTEXT_ATTRIBUTE_NAMES"],
                        "FIRST_CONTEXT_PATH": nD["CONTEXT_ATTRIBUTE_NAMES"][0],
                        "CONTEXT_ATTRIBUTE_VALUES": cVDL,
                    }
                else:
                    subCatD[(nD["CATEGORY"], nD["SUBCATEGORY"])] = {"CONTEXT_NAME": nD["SUBCATEGORY"]}
            cD[collectionName] = subCatD
        return cD

    def isSubCategoryNested(self, collectionName, categoryName, subCategoryName):
        """Return is the input subcategory in this collection is nested.

        Args:
            collectionName (str): collection name
            categoryName (str): category name
            subCategoryName (str): subcategory name

        Returns:
            (bool): True if nested or False otherwise

        """
        ret = False
        try:
            self.__subCategoryNested = self.__prepareSubCategoryNested() if not self.__subCategoryNested else self.__subCategoryNested
            ret = (categoryName, subCategoryName) in self.__subCategoryNested[collectionName]
        except Exception:
            pass
        return ret

    def getSubCategoryNestedContext(self, collectionName, categoryName, subCategoryName):
        """Return is the input subcategory in this collection is nested.

        Args:
            collectionName (str): collection name
            categoryName (str): categoryName
            subCategoryName (str): subcategory name

        Returns:
            (dict): {"CONTEXT_NAME": <name>, "CONTEXT_PATHS": <full_path_list>}

        """
        ret = None
        try:
            self.__subCategoryNested = self.__prepareSubCategoryNested() if not self.__subCategoryNested else self.__subCategoryNested
            ret = self.__subCategoryNested[collectionName][(categoryName, subCategoryName)]
        except Exception:
            pass
        logger.debug("collection %r category %r subcat %r : %r", collectionName, categoryName, subCategoryName, ret)
        return ret

    def getAllSubCategoryNestedContexts(self):
        """Return is the full nested subcategory data structure.

        Args:
            categoryName (str): categoryName
            subCategoryName (str): subcategory name

        Returns:
            (dict): {"CONTEXT_NAME": <name>, "CONTEXT_PATHS": <full_path_list>}

        """
        ret = None
        try:
            self.__subCategoryNested = self.__prepareSubCategoryNested() if not self.__subCategoryNested else self.__subCategoryNested
            ret = self.__subCategoryNested
        except Exception:
            pass
        return ret

    def __prepareAttributeSearchPriorities(self):
        """
        Example:

            collection_attribute_search_priority:
                pdbx_core_entry:
                    - ATTRIBUTE_NAME: rcsb_entry_container_identifiers.entry_id
                    PRIORITY_VALUE: 20
                    - ATTRIBUTE_NAME: entity.rcsb_macromolecular_names_combined
                    PRIORITY_VALUE: 20

        """
        pD = {}
        try:
            # preprocess priority data --
            for collectionName, tDL in self.__cfgD["collection_attribute_search_priority"].items():
                aD = {}
                for tD in tDL:
                    ff = str(tD["ATTRIBUTE_NAME"]).split(".")
                    pValue = tD["PRIORITY_VALUE"]
                    aD[(ff[0], ff[1])] = pValue
                pD[collectionName] = aD
            return pD
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return pD

    def getAttributeTextSearchPriority(self, collectionName, categoryName, attributeName):
        try:
            self.__attributeSeachPriority = self.__prepareAttributeSearchPriorities() if not self.__attributeSeachPriority else self.__attributeSeachPriority
            try:
                # return a config priority first
                return self.__attributeSeachPriority[collectionName](categoryName, attributeName)
            except Exception:
                pass
            # return an elevated priority based on search context -
            #
            scL = self.getAttributeSearchContexts(collectionName, categoryName, attributeName)
            if "suggest" in scL:
                return 20
            if "exact-match" in scL:
                return 10
            if "full-text" in scL:
                return 1
        except Exception:
            pass
        return None

    def __prepareAttributeSearchGroups(self):
        """
        Example:

        search_group_membership:
            - GROUP_NAME: ID(s) and Keywords
              ATTRIBUTE_NAME_LIST:
                - rcsb_entry_container_identifiers.entry_id
                - pdbx_deposit_group.group_id
                - rcsb_pubmed_container_identifiers.pubmed_id
                - rcsb_polymer_entity_container_identifiers.reference_sequence_identifiers.database_accession
                - rcsb_polymer_entity_container_identifiers.reference_sequence_identifiers.database_name
                - struct_keywords.pdbx_keywords
                - rcsb_entity_source_organism.rcsb_gene_name.value
            - GROUP_NAME: Structure Annotation
              ATTRIBUTE_NAME_LIST:
                - rcsb_polymer_entity.pdbx_description
                - rcsb_polymer_entity.rcsb_macromolecular_names_combined.name
                - rcsb_membrane_lineage.name
                - pdbx_database_status.pdb_format_compatible

        returns:
            dict : {(category, attribute): [(search group, iorder), ...], }, }
            dict : {group_name: [(category, attribute), ... ], ...}
        """
        aD = {}
        gD = {}
        try:
            for sgD in self.__cfgD["search_group_membership"]:
                groupName = sgD["GROUP_NAME"]
                jj = 5
                for atName in sgD["ATTRIBUTE_NAMES"]:
                    ff = atName.split(".")
                    if len(ff) != 2:
                        logger.error("Bad attribute name for search group %r", atName)
                        continue
                    aD.setdefault((ff[0], ff[1]), []).append((groupName, jj))
                    gD.setdefault(groupName, []).append((ff[0], ff[1]))
                    jj += 5

        except Exception as e:
            logger.exception("Failing with sgD %r %s", sgD if sgD else None, str(e))
        return gD, aD

    def inSearchGroup(self, categoryName, attributeName):
        """[summary]

        Args:
            categoryName ([type]): [description]
            attributeName ([type]): [description]

        Returns:
            bool: True if in a search group or False otherwise
        """
        try:
            self.__searchGroupD, self.__searchGroupAttributeD = (
                self.__prepareAttributeSearchGroups() if not self.__searchGroupD else (self.__searchGroupD, self.__searchGroupAttributeD)
            )
            return self.__searchGroupAttributeD.get((categoryName, attributeName), None) is not None
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def getSearchGroup(self, categoryName, attributeName):
        """[summary]

        Args:
            categoryName ([type]): [description]
            attributeName ([type]): [description]

        Returns:
            list: [(groupName, priorityOrder),...]
        """
        try:
            self.__searchGroupD, self.__searchGroupAttributeD = (
                self.__prepareAttributeSearchGroups() if not self.__searchGroupD else (self.__searchGroupD, self.__searchGroupAttributeD)
            )
            return self.__searchGroupAttributeD.get((categoryName, attributeName), [])
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return []

    def getSearchGroupAttributes(self, searchGroupName):
        try:
            self.__searchGroupD, self.__searchGroupAttributeD = (
                self.__prepareAttributeSearchGroups() if not self.__searchGroupD else (self.__searchGroupD, self.__searchGroupAttributeD)
            )
            return self.__searchGroupD.get(searchGroupName, [])
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return []

    def getSearchGroups(self):
        try:
            self.__searchGroupD, self.__searchGroupAttributeD = (
                self.__prepareAttributeSearchGroups() if not self.__searchGroupD else (self.__searchGroupD, self.__searchGroupAttributeD)
            )
            return list(self.__searchGroupD.keys())
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return []

    def checkSearchGroups(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        groupNameList = self.getSearchGroups()
        logger.info("Search groups (%d)", len(groupNameList))
        for groupName in groupNameList:
            # get attributes in group
            attributeTupList = self.getSearchGroupAttributes(groupName)
            logger.info("Search Group (%2d): %s", len(attributeTupList), groupName)
            # get search context and brief descriptions -
            for catName, atName in attributeTupList:
                searchContextTupL = self.getSearchContexts(catName, atName)
                if not searchContextTupL:
                    logger.warning("Missing search context for %s.%s", catName, atName)
                descriptionText = self.getAttributeDescription(catName, atName, contextType="brief")
                if not descriptionText:
                    logger.warning("Missing brief description %s.%s", catName, atName)
                #
                nestedContextDL = self.getNestedContexts(catName)
                for nestedContextD in nestedContextDL:
                    contextName = nestedContextD["CONTEXT_NAME"]
                    contextPath = nestedContextD["CONTEXT_PATH"] if "CONTEXT_PATH" in nestedContextD else None
                    if contextPath:
                        cpCatName = contextPath.split(".")[0]
                        cpAtName = contextPath.split(".")[1]
                        nestedPathSearchContext = self.getSearchContexts(cpCatName, cpAtName)
                        if not nestedPathSearchContext:
                            logger.warning("Missing nested (%r) search context for %r %r", contextName, cpCatName, cpAtName)
                #
                logger.debug("  %r %r -> %r (%s)", catName, atName, descriptionText, ",".join([tup[0] for tup in searchContextTupL]))
        return True
